import opengate as gate
import opengate_core as g4


class ConsVolume(gate.VolumeBase):
    """
    http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html
    """

    type_name = "Cons"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        # default values
        u = user_info
        mm = gate.g4_units("mm")
        deg = gate.g4_units("deg")
        u.rmin1 = 5 * mm
        u.rmax1 = 10 * mm
        u.rmin2 = 20 * mm
        u.rmax2 = 25 * mm
        u.dz = 40 * mm
        u.sphi = 0 * deg
        u.dphi = 45 * deg

    def build_solid(self):
        u = self.user_info
        return g4.G4Cons(
            u.name, u.rmin1, u.rmax1, u.rmin2, u.rmax2, u.dz, u.sphi, u.dphi
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test039_hits_memory_check_base import *

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = create_simu(8)
ui = sim.user_info
ui.random_seed = "auto"

# go
sim.initialize()
sim.start()

is_ok = test_results(sim)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
import opengate.contrib.phantom_nema_iec_body as gate_iec
import gatetools.phsp as phsp
import uproot
import numpy as np
import matplotlib.pyplot as plt
import os


def create_simulation(sim, paths):
    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    cm3 = gate.g4_units("cm3")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    BqmL = Bq / cm3

    # main parameters
    ui = sim.user_info
    ui.check_volumes_overlap = True
    ui.random_seed = 123456
    # ac = 1e6 * BqmL
    ac = 3e3 * BqmL / ui.number_of_threads
    ui.visu = False

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # iec phantom not needed
    # iec_phantom = gate_iec.add_phantom(sim)

    # cylinder of the phase space, for visualisation only
    """cyl = sim.add_volume('Sphere', 'phase_space_cylinder')
    cyl.rmin = 210 * mm
    cyl.rmax = 211 * mm
    cyl.color = [1, 1, 1, 1]
    cyl.material = 'G4_AIR'"""

    # test phase space to check with reference
    phase_space_sphere = sim.add_volume("Sphere", "phase_space_sphere")
    phase_space_sphere.rmin = 212 * mm
    phase_space_sphere.rmax = 213 * mm
    phase_space_sphere.color = [1, 1, 1, 1]
    phase_space_sphere.material = "G4_AIR"

    # spect head
    distance = 30 * cm
    psd = 6.11 * cm
    p = [0, 0, -(distance + psd)]
    spect1 = gate_spect.add_ge_nm67_spect_head(
        sim, "spect1", collimator_type="lehr", debug=ui.visu
    )
    spect1.translation, spect1.rotation = gate.get_transform_orbiting(p, "x", 180)

    # spect head (debug mode = very small collimator)
    # spect2 = gate_spect.add_ge_nm67_spect_head(sim, 'spect2', collimator=colli_flag, debug=False)
    # spect2.translation, spect2.rotation = gate.get_transform_orbiting(p, 'x', 0)

    # physic list
    sim.set_cut("world", "all", 1 * mm)

    # activity parameters
    spheres_diam = [10, 13, 17, 22, 28, 37]
    spheres_activity_concentration = [ac * 6, ac * 5, ac * 4, ac * 3, ac * 2, ac]
    # spheres_diam = [37]
    # spheres_activity_concentration = [ac] * len(spheres_diam)

    # initialisation for conditional
    spheres_radius = [x / 2.0 for x in spheres_diam]
    # spheres_centers, spheres_volumes = gate_iec.compute_sphere_centers_and_volumes(sim, iec_phantom.name)
    spheres_centers, spheres_volumes = gate_iec.get_default_sphere_centers_and_volumes()
    spheres_activity_ratio = []
    spheres_activity = []
    for diam, ac, volume, center in zip(
        spheres_diam, spheres_activity_concentration, spheres_volumes, spheres_centers
    ):
        activity = ac * volume
        print(
            f"Sphere {diam}: {str(center):<30} {volume / cm3:7.3f} cm3 "
            f"{activity / Bq:7.0f} Bq  {ac / BqmL:7.1f} BqmL"
        )
        spheres_activity.append(activity)

    total_activity = sum(spheres_activity)
    print(f"Total activity {total_activity / Bq:.0f} Bq")
    for activity in spheres_activity:
        spheres_activity_ratio.append(activity / total_activity)
    print("Activity ratio ", spheres_activity_ratio, sum(spheres_activity_ratio))
    # print('Radius ', spheres_radius)
    # print('Volumes ', spheres_volumes)

    # unique (reproducible) random generator
    rs = gate.get_rnd_seed(123456)

    class GANTest:
        def __init__(self):
            # will store all conditional info (position, direction)
            # (not needed, only for test)
            self.all_cond = None

        def generate_condition(self, n):
            n_samples = gate_iec.get_n_samples_from_ratio(n, spheres_activity_ratio)
            cond = gate_iec.generate_pos_dir_spheres(
                spheres_centers, spheres_radius, n_samples, shuffle=True, rs=rs
            )

            if self.all_cond is None:
                self.all_cond = cond
            else:
                self.all_cond = np.column_stack((self.all_cond, cond))

            return cond

    # GAN source
    gsource = sim.add_source("GAN", "gaga")
    gsource.particle = "gamma"
    # no phantom, we consider attached to the world at origin
    # gsource.mother = f'{iec_phantom.name}_interior'
    gsource.activity = total_activity
    gsource.pth_filename = paths.gate / "pth2" / "test001_GP_0GP_10_50000.pth"
    gsource.position_keys = ["PrePosition_X", "PrePosition_Y", "PrePosition_Z"]
    gsource.backward_distance = 5 * cm
    gsource.direction_keys = ["PreDirection_X", "PreDirection_Y", "PreDirection_Z"]
    gsource.energy_key = "KineticEnergy"
    gsource.energy_threshold = 0.001 * keV
    gsource.weight_key = None
    gsource.time_key = "TimeFromBeginOfEvent"
    gsource.time_relative = True
    gsource.batch_size = 5e4
    gsource.verbose_generator = True

    # GANSourceConditionalGenerator manages the conditional GAN
    # GANTest manages the generation of the conditions, we use a class here to store the total
    # list of conditions (only needed for the test)
    condition_generator = GANTest()
    gen = gate.GANSourceConditionalGenerator(
        gsource, condition_generator.generate_condition
    )
    gsource.generator = gen

    # it is possible to use acceptance angle. Not done here to check exiting phsp
    # gsource.direction.acceptance_angle.volumes = [spect1.name]
    # gsource.direction.acceptance_angle.intersection_flag = True

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.output = paths.output / "test038_gan_stats.txt"

    # add default digitizer (it is easy to change parameters if needed)
    gate_spect.add_simplified_digitizer_Tc99m(
        sim, "spect1_crystal", paths.output / "test038_gan_proj.mhd"
    )
    # gate_spect.add_ge_nm670_spect_simplified_digitizer(sim, 'spect2_crystal', paths.output / 'test033_proj_2.mhd')
    singles_actor = sim.get_actor_user_info(f"Singles_spect1_crystal")
    singles_actor.output = paths.output / "test038_gan_singles.root"

    # motion of the spect, create also the run time interval
    """heads = [spect1]  # [spect1, spect2]

    # create a list of run (total = 1 second)
    n = 1
    sim.run_timing_intervals = gate.range_timing(0, 1 * sec, n)

    for head in heads:
        motion = sim.add_actor('MotionVolumeActor', f'Move_{head.name}')
        motion.mother = head.name
        motion.translations, motion.rotations = \
            gate.volume_orbiting_transform('x', 0, 180, n, head.translation, head.rotation)
        motion.priority = 5"""

    phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp_actor.mother = phase_space_sphere.name
    phsp_actor.attributes = [
        "KineticEnergy",
        "PrePosition",
        "PreDirection",
        "GlobalTime",
        "EventPosition",
        "EventDirection",
        "EventKineticEnergy",
    ]
    phsp_actor.output = paths.output / "test038_gan_phsp.root"

    return condition_generator


def analyze_results(sim, paths, all_cond):
    ui = sim.user_info
    gsource = sim.get_source_user_info("gaga")
    phsp_actor = sim.get_actor_user_info("phsp")

    # print stats
    print()
    gate.warning(f"Check stats")
    if ui.number_of_threads == 1:
        s = sim.get_source("gaga")
    else:
        s = sim.get_source_MT("gaga", 0)
    print(f"Source, nb of skipped particles (absorbed) : {s.fNumberOfSkippedParticles}")
    b = gate.get_source_skipped_particles(sim, gsource.name)
    print(f"Source, nb of skipped particles (AA)       : {b}")

    stats = sim.get_actor("Stats")
    print(stats)
    stats_ref = gate.read_stat_file(paths.output_ref / "test038_ref_stats.txt")
    r = (
        stats_ref.counts.step_count - stats.counts.step_count
    ) / stats_ref.counts.step_count
    print(f"!!! Steps cannot be compared => was {stats.counts.step_count}, {r:.2f}%")
    stats.counts.step_count = stats_ref.counts.step_count
    stats.counts.run_count = 1  # force for MT
    is_ok = gate.assert_stats(stats, stats_ref, 0.10)

    # save conditional for checking with reference cond
    keys = [
        "EventPosition_X",
        "EventPosition_Y",
        "EventPosition_Z",
        "EventDirection_X",
        "EventDirection_Y",
        "EventDirection_Z",
    ]
    phsp.save_npy(paths.output / "test038_gan_phsp_cond.npy", all_cond, keys)

    # ----------------------------------------------------------------------------------------------
    # compare conditional
    # less particle in the ref because conditional data are stored
    # when exit (not absorbed)
    print()
    gate.warning(f"Check conditions (position, direction)")
    root_ref = (
        paths.output_ref / "test038_ref_phsp.root"
    )  # looking the previous generated
    hits1 = uproot.open(root_ref)
    branch = hits1.keys()[0]
    print("Branch name:", branch)
    hits1 = hits1[branch]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")
    root_gan = paths.output / "test038_gan_phsp_cond.npy"
    hits2, hits2_keys, hits2_n = phsp.load(root_gan)
    tols = [10.0] * len(keys)
    tols[keys.index("EventPosition_X")] = 0.3
    tols[keys.index("EventPosition_Y")] = 0.5
    tols[keys.index("EventPosition_Z")] = 0.3
    tols[keys.index("EventDirection_X")] = 0.02
    tols[keys.index("EventDirection_Y")] = 0.02
    tols[keys.index("EventDirection_Z")] = 0.02
    scalings = [1] * len(keys)
    is_ok = (
        gate.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2_keys),
            keys,
            keys,
            tols,
            scalings,
            scalings,
            True,
        )
        and is_ok
    )
    # figure
    img_filename = paths.output / "test038_cond.png"
    plt.suptitle(
        f"Values: ref {os.path.basename(root_ref)} {os.path.basename(root_gan)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img_filename)
    print(f"Figure in {img_filename}")

    # ----------------------------------------------------------------------------------------------
    # compare output phsp
    """
        This is *not* a very good pth for the moment, we set a high tolerance.
    """
    print()
    gate.warning(f"Check output phsp")
    ref_file = paths.output_ref / "test038_ref_phsp.root"
    hc_file = phsp_actor.output
    checked_keys = [
        "GlobalTime",
        "KineticEnergy",
        "PrePosition_X",
        "PrePosition_Y",
        "PrePosition_Z",
        "PreDirection_X",
        "PreDirection_Y",
        "PreDirection_Z",
    ]
    scalings = [1.0] * len(checked_keys)
    scalings[checked_keys.index("GlobalTime")] = 1e-9  # time in ns
    tols = [10.0] * len(checked_keys)
    tols[checked_keys.index("GlobalTime")] = 0.003
    tols[checked_keys.index("KineticEnergy")] = 0.002
    tols[checked_keys.index("PrePosition_X")] = 7
    tols[checked_keys.index("PrePosition_Y")] = 4
    tols[checked_keys.index("PrePosition_Z")] = 4
    tols[checked_keys.index("PreDirection_X")] = 0.02
    tols[checked_keys.index("PreDirection_Y")] = 0.02
    tols[checked_keys.index("PreDirection_Z")] = 0.02
    print(scalings, tols)
    is_ok = (
        gate.compare_root3(
            ref_file,
            hc_file,
            "phsp",
            "phsp",
            checked_keys,
            checked_keys,
            tols,
            scalings,
            scalings,
            paths.output / "test038_phsp.png",
        )
        and is_ok
    )

    # ----------------------------------------------------------------------------------------------
    # compare hits
    print()
    gate.warning(f"Check singles -> NOT YET (too low statistics)")

    """ref_file = paths.output / 'test038_ref_singles.root'
    hc_file = singles_actor.output
    checked_keys = ['GlobalTime', 'TotalEnergyDeposit', 'PostPosition_X', 'PostPosition_Y', 'PostPosition_Z']
    scalings = [1.0] * len(checked_keys)
    scalings[checked_keys.index('GlobalTime')] = 1e-9  # time in ns
    tols[checked_keys.index('GlobalTime')] = 0.2
    tols[checked_keys.index('TotalEnergyDeposit')] = 10
    tols[checked_keys.index('PostPosition_X')] = 100
    tols[checked_keys.index('PostPosition_Y')] = 100
    tols[checked_keys.index('PostPosition_Z')] = 100
    print(scalings, tols)
    is_ok = gate.compare_root3(ref_file, hc_file, "Singles_spect1_crystal", "Singles_spect1_crystal",
                              checked_keys, checked_keys, tols, scalings, scalings,
                              paths.output / 'test038_singles.png', hits_tol=100) and is_ok
    """
    # ----------------------------------------------------------------------------------------------

    # this is the end, my friend
    # gate.delete_run_manager_if_needed(sim)
    gate.test_ok(is_ok)



from .GenericSource import *
import sys
import time
import scipy
import threading


class GANSourceDefaultGenerator:
    """
    This class manage the base components of a particle generator.

    - In the constructor, the module 'gaga' is imported. It is only imported in the constructor to only required
    this module if it is used

    - 'initialize' function: the GAN is loaded and the list of keys is initialized

    - 'generator' function: default generator

    - 'get_output_keys' function: map the user defined keys to the ones of the generator. There are two usages, either
    with on single primary (3 values for position, direction), or paired primary (6 values).

    - 'move_backward' function: consider all particles positions and move their backward according to their direction,
    using the factor provided by the user in 'user_info.backward_distance'. This is useful to allow generating
    particles that do not intersect with the detector.

    - 'copy_generated_particle_to_g4' function: copy all the particles (pos, dir, time, energy) to the cpp part.

    """

    def __init__(self, user_info):
        self.user_info = user_info
        self.gaga = gate.import_gaga_phsp()
        self.indexes_are_build = None
        if self.gaga is None:
            print("Cannot run GANSource")
            sys.exit()
        self.lock = threading.Lock()
        self.initialize_is_done = False
        self.keys_output = None

    def initialize(self):
        with self.lock:
            if not self.initialize_is_done:
                self.initialize_with_lock()
                self.initialize_is_done = True

    def initialize_with_lock(self):
        # allow converting str like 1e5 to int
        self.user_info.batch_size = int(float(self.user_info.batch_size))

        # read pth
        self.gan = Box()
        g = self.gan
        g.params, g.G, g.D, g.optim, g.dtypef = self.gaga.load(
            self.user_info.pth_filename, "auto", verbose=False
        )

        # by default, the output keys are the same as the input keys
        # (could be changed later by some parameterization)
        if self.keys_output is None:
            g.params.keys_output = g.params.keys_list
        else:
            g.params.keys_output = self.keys_output

        # check the number of params
        dim = len(self.user_info.position_keys)
        if dim != 3 and dim != 6:
            gate.fatal(
                f"In the source {self.user_info.name}, "
                f"position_keys size must be 3 or 6, while it is {dim}"
            )
        dim2 = len(self.user_info.direction_keys)
        if dim2 != dim:
            gate.fatal(
                f"In the source {self.user_info.name}, "
                f"direction_keys must have the same size as position_keys, while it is {dim2} and {dim}"
            )
        # if 3 only (not paired generation), the E, w and t are put in a vector
        if dim == 3:
            self.user_info.is_paired = False
            self.user_info.energy_key = [self.user_info.energy_key]
            if self.user_info.weight_key is not None:
                self.user_info.weight_key = [self.user_info.weight_key]
            if self.user_info.time_key is not None:
                self.user_info.time_key = [self.user_info.time_key]
        else:
            self.user_info.is_paired = True

    def get_output_keys(self):
        with self.lock:
            # (this is only performed once)
            if not self.indexes_are_build:
                self.get_output_keys_with_lock()
                self.indexes_are_build = True

    def get_output_keys_with_lock(self):
        g = self.gan

        # get position index from GAN (or a fixed value)
        k = g.params.keys_output
        print("k", k)
        n = self.user_info.batch_size
        dim = len(self.user_info.position_keys)
        g.position, g.position_type = self.get_key_generated_values(
            k, self.user_info.position_keys, n, dim=dim
        )

        # get position from GAN (or a fixed value)
        g.direction, g.direction_type = self.get_key_generated_values(
            k, self.user_info.direction_keys, n, dim=dim
        )

        # one primary or two primaries (paired) ?
        d = 1
        if self.user_info.is_paired:
            d = 2

        # get energy from GAN (or a fixed value)
        g.energy, g.energy_type = self.get_key_generated_values(
            k, self.user_info.energy_key, n, dim=d
        )

        # get weight from GAN (or a fixed value)
        if self.user_info.weight_key is not None:
            g.weight, g.weight_type = self.get_key_generated_values(
                k, self.user_info.weight_key, n, dim=d
            )

        # get time from GAN (or a fixed value)
        if self.user_info.time_key is not None:
            g.time, g.time_type = self.get_key_generated_values(
                k, self.user_info.time_key, n, dim=d
            )

    def get_key_generated_values(self, k, pk, n, dim=3):
        p = []
        o = []  # true or false if used or not
        for i in range(dim):
            try:
                index = k.index(pk[i])
                p.append(index)
                o.append(True)
            except:
                if type(pk[i]) == float:
                    p.append(np.ones(n) * pk[i])
                    o.append(False)
                else:
                    gate.fatal(
                        f"Error, cannot use the key {pk[i]} (in {pk}) in the GAN source. "
                        f"GAN keys are: {k}"
                    )
        return p, o

    def generator(self, source):
        """
        Main function that will be called from the cpp side every time a batch
        of particles should be created.
        Once created here, the particles are copied to cpp.
        (Yes maybe the copy could be avoided, but I did not manage to do it)
        """
        # get the info
        g = self.gan
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ...", end="")
            sys.stdout.flush()

        # generate samples (this is the most time-consuming part)
        fake = self.gaga.generate_samples2(
            g.params,
            g.G,
            g.D,
            n=n,
            batch_size=n,
            normalize=False,
            to_numpy=True,
            silence=True,
        )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f" done in {end - start:0.1f} sec (GPU={g.params.current_gpu})")

    def copy_generated_particle_to_g4(self, source, g, fake):
        # get the values from GAN or fixed value
        # the index are precomputed in get_key_generated_values
        # (this is a bit convoluted, but it does the job)
        pos = []
        dir = []
        dim = len(g.position)
        is_paired = self.user_info.is_paired

        # position, direction
        for i in range(dim):
            if g.position_type[i]:
                pos.append(fake[:, g.position[i]])
            else:
                pos.append(g.position[i])
            if g.direction_type[i]:
                dir.append(fake[:, g.direction[i]])
            else:
                dir.append(g.direction[i])

        # energy
        if g.energy_type:
            energy = []
            energy.append(fake[:, g.energy[0]])
            if is_paired:
                energy.append(fake[:, g.energy[1]])
        else:
            energy = g.energy

        # weight ? (fake if not used)
        weight = [0]
        if self.user_info.weight_key is not None:
            weight = []
            weight.append(fake[:, g.weight[0]])
            if is_paired:
                weight.append(fake[:, g.weight[1]])

        # time (fake if not used)
        the_time = [0]
        if self.user_info.time_key is not None:
            the_time = []
            the_time.append(fake[:, g.time[0]])
            if is_paired:
                the_time.append(fake[:, g.time[1]])

        # copy to c++
        source.fPositionX = pos[0]
        source.fPositionY = pos[1]
        source.fPositionZ = pos[2]

        # copy to c++
        source.fDirectionX = dir[0]
        source.fDirectionY = dir[1]
        source.fDirectionZ = dir[2]

        # copy to c++
        source.fEnergy = energy[0]
        if self.user_info.weight_key is not None:
            source.fWeight = weight[0]
        if self.user_info.time_key is not None:
            source.fTime = the_time[0]

        # pairs ?
        if is_paired:
            source.fPositionX2 = pos[3]
            source.fPositionY2 = pos[4]
            source.fPositionZ2 = pos[5]

            # copy to c++
            source.fDirectionX2 = dir[3]
            source.fDirectionY2 = dir[4]
            source.fDirectionZ2 = dir[5]

            # copy to c++
            source.fEnergy2 = energy[1]
            if self.user_info.weight_key is not None:
                source.fWeight2 = weight[1]
            if self.user_info.time_key is not None:
                source.fTime2 = the_time[1]

    def move_backward(self, g, fake):
        # move particle backward ?
        back = self.user_info.backward_distance
        if not back:
            return
        # move particle position
        position = fake[:, g.position[0] : g.position[0] + 3]
        direction = fake[:, g.direction[0] : g.direction[0] + 3]
        fake[:, g.position[0] : g.position[0] + 3] = position - back * direction

        # modify the time because we move the particle backward
        c = scipy.constants.speed_of_light * 1000 / 1e9  # in mm/ns
        xt = back / c
        if self.user_info.time_key is not None:
            fake[:, g.time[0]] -= xt

        if self.user_info.is_paired:
            # move second particle position
            position = fake[:, g.position[3] : g.position[3] + 3]
            direction = fake[:, g.direction[3] : g.direction[3] + 3]
            fake[:, g.position[3] : g.position[3] + 3] = position - back * direction
            if self.user_info.time_key is not None:
                fake[:, g.time[1]] -= xt



import opengate as gate


class SourceBase(gate.UserElement):
    """
    Base class for all source types.
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        # user properties shared by all sources
        user_info.mother = gate.__world_name__
        user_info.start_time = None
        user_info.end_time = None

    def __init__(self, user_info):
        # type_name MUST be defined in class that inherit from SourceBase
        super().__init__(user_info)
        # the cpp counterpart of the source
        self.g4_source = self.create_g4_source()
        # all times intervals
        self.run_timing_intervals = None

    def __str__(self):
        s = f"{self.user_info.name}: {self.user_info}"
        return s

    def dump(self):
        sec = gate.g4_units("s")
        start = "no start time"
        end = "no end time"
        if self.user_info.start_time is not None:
            start = f"{self.user_info.start_time / sec} sec"
        if self.user_info.end_time is not None:
            end = f"{self.user_info.end_time / sec} sec"
        s = (
            f"Source name        : {self.user_info.physics_list_name}\n"
            f"Source type        : {self.user_info.type}\n"
            f"Start time         : {start}\n"
            f"End time           : {end}"
        )
        return s

    def __del__(self):
        pass

    def create_g4_source(self):
        gate.fatal('The function "create_g4_source" *must* be overridden')

    def initialize(self, run_timing_intervals):
        self.run_timing_intervals = run_timing_intervals
        # by default consider the source time start and end like the whole simulation
        # Start: start time of the first run
        # End: end time of the last run
        if not self.user_info.start_time:
            self.user_info.start_time = run_timing_intervals[0][0]
        if not self.user_info.end_time:
            self.user_info.end_time = run_timing_intervals[-1][1]
        # this will initialize and set user_info to the cpp side
        self.g4_source.InitializeUserInfo(self.user_info.__dict__)

    def get_estimated_number_of_events(self, run_timing_interval):
        gate.fatal(f"Not implemented yet: get_estimated_number_of_events")
        exit()
        # by default, all event have the same time, so we check that
        # this time is included into the given time interval
        if (
            run_timing_interval[0]
            <= self.user_info.start_time
            <= run_timing_interval[1]
        ):
            return self.user_info.n
        return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test043_garf_helpers import *
import opengate.contrib.spect_ge_nm670 as gate_spect
import math

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 123654
ui.number_of_threads = 1

# units
activity = 1 * Bq

# world
sim_set_world(sim)

# distance
spect_translation = 50 * cm
debug = True

# spect head n°1
spect1 = gate_spect.add_ge_nm67_spect_head(sim, "spect_lehr", "lehr", debug=debug)
spect1.translation = [0, 0, -spect_translation]

# spect head n°2
spect2 = gate_spect.add_ge_nm67_spect_head(sim, "spect_megp", "megp", debug=debug)
p = [0, 0, -spect_translation]
itr, irot = gate.get_transform_orbiting(p, "x", 180)
spect2.translation = itr
spect2.rotation = irot

# spect head n°3
spect3 = gate_spect.add_ge_nm67_spect_head(sim, "spect_hegp", "hegp", debug=debug)
p = [0, 0, -spect_translation]
itr, irot = gate.get_transform_orbiting(p, "x", 90)
spect3.translation = itr
spect3.rotation = irot

# physics
sim_phys(sim)

# sources
sim_source_test(sim, activity)

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True

# create planes only to check visually (volume overlap of course)
print()
print(f"Distances are from the center of the head volume, in the Z direction")

collis = ["lehr", "megp", "hegp"]
plane_positions = {}
distance_to_crystal = {}
psd_dist = {}
for colli in collis:
    name = f"spect_{colli}"
    x = gate_spect.get_volume_position_in_head(sim, name, "collimator_psd", "max")
    plane_positions[colli] = x
    print(f"{colli} PSD distance         : {x / mm} mm")
    a = spect1.size[2] / 2.0 - x
    psd_dist[colli] = a
    print(f"{colli} PSD distance from BB : {a / mm} mm")
    detPlane1 = sim_add_detector_plane(sim, name, x, f"dp_psd_{colli}")
    # for visualization purpose only we increase the size by 20%
    detPlane1.size[0] *= 1.2
    detPlane1.size[1] *= 1.2
    y = gate_spect.get_volume_position_in_head(sim, name, "crystal", "center")
    print(f"{colli} crystal distance     : {y / mm} mm")
    detPlane2 = sim_add_detector_plane(sim, name, y, f"dp_crystal_{colli}")
    # for visualization purpose only we increase the size by 20%
    detPlane2.size[0] *= 1.2
    detPlane2.size[1] *= 1.2
    print(f"{colli} crystal to PSD dist  : {(x - y / mm)} mm")
    distance_to_crystal[colli] = x - y
    print()

# --------------------------------------------------------------------------------
# create G4 objects only if visu (no need for the test)
if ui.visu:
    ui.verbose_level = gate.NONE
    ui.check_volumes_overlap = False
    sim.initialize()

    # start simulation
    sim.start()

# --------------------------------------------------------------------------------
# check values
is_ok = True
for colli in collis:
    pp, dc, psd = gate_spect.get_plane_position_and_distance_to_crystal(colli)
    ok = math.isclose(pp, plane_positions[colli])
    gate.print_test(
        ok,
        f"Colli {colli} detector plane position       : {pp:5.2f}  vs  {plane_positions[colli]:5.2f} mm",
    )
    is_ok = is_ok and ok
    ok = math.isclose(dc, distance_to_crystal[colli])
    gate.print_test(
        ok,
        f"Colli {colli} distance to crystal           : {dc:5.2f}  vs  {distance_to_crystal[colli]:5.2f} mm",
    )
    is_ok = is_ok and ok
    ok = math.isclose(psd, psd_dist[colli])
    gate.print_test(
        ok,
        f"Colli {colli} distance head boundary to PSD : {psd:5.2f}  vs  {psd_dist[colli]:5.2f} mm",
    )
    is_ok = is_ok and ok

print()
print(
    "Warning : with simulation, you should add 1 nm to the position to avoid overlap."
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import pathlib

current_path = pathlib.Path(__file__).parent.resolve()
data_path = current_path / ".." / "data"
ref_path = current_path / ".." / "data" / "gate" / "gate_test008_dose_actor" / "output"
output_path = current_path / ".." / "output"

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = "auto"

#  change world size
m = gate.g4_units("m")
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a simple fake volume to test hierarchy
# translation and rotation like in the Gate macro
fake = sim.add_volume("Box", "fake")
cm = gate.g4_units("cm")
fake.size = [40 * cm, 40 * cm, 40 * cm]
fake.translation = [1 * cm, 2 * cm, 3 * cm]
fake.rotation = Rotation.from_euler("x", 10, degrees=True).as_matrix()
fake.material = "G4_AIR"
fake.color = [1, 0, 1, 1]

# waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.mother = "fake"
waterbox.size = [10 * cm, 10 * cm, 10 * cm]
waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
waterbox.rotation = Rotation.from_euler("y", 20, degrees=True).as_matrix()
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMV"
p.enable_decay = False
p.apply_cuts = True  # default
cuts = p.production_cuts
um = gate.g4_units("um")
cuts.world.gamma = 700 * um
cuts.world.electron = 700 * um
cuts.world.positron = 700 * um
cuts.world.proton = 700 * um

# default source for tests
source = sim.add_source("Generic", "mysource")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.energy.mono = 150 * MeV
nm = gate.g4_units("nm")
source.particle = "proton"
source.position.type = "disc"
source.position.radius = 1 * nm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 50000 * Bq

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = output_path / "test008-edep.mhd"
dose.mother = "waterbox"
dose.size = [99, 99, 99]
mm = gate.g4_units("mm")
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.translation = [2 * mm, 3 * mm, -2 * mm]
dose.uncertainty = True
dose.hit_type = "random"

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# print results at the end
stat = sim.get_actor("Stats")
print(stat)

dose = sim.get_actor("dose")
print(dose)

# tests
stats_ref = gate.read_stat_file(ref_path / "stat.txt")
is_ok = gate.assert_stats(stat, stats_ref, 0.10)

print("\nDifference for EDEP")
is_ok = (
    gate.assert_images(
        ref_path / "output-Edep.mhd",
        output_path / "test008-edep.mhd",
        stat,
        tolerance=13,
        ignore_value=0,
    )
    and is_ok
)

print("\nDifference for uncertainty")
is_ok = (
    gate.assert_images(
        ref_path / "output-Edep-Uncertainty.mhd",
        output_path / "test008-edep_uncertainty.mhd",
        stat,
        tolerance=30,
        ignore_value=1,
    )
    and is_ok
)

gate.test_ok(is_ok)



# for a unclear reason, ssl must be imported before to avoid error:
# "from http.client import HTTPConnection, HTTPSConnection
# ImportError: cannot import name 'HTTPSConnection' from 'http.client'"
import ssl

# generic helpers
from .geometry.VolumeManager import __world_name__
from .helpers_log import *
from .helpers import *
from .helpers_image import *
from .helpers_tests import *
from .helpers_tests_root import *
from .helpers_transform import *

# main mechanism for the 'elements': source, actor, volume
from .UserInfo import *
from .UserElement import *
from .source.SourceBase import *
from .actor.ActorBase import *
from .geometry.VolumeBase import *

# main object
from .Simulation import *
from .helpers_run_timing import *

# helpers to list all possible types of elements
from .geometry.helpers_geometry import *
from .geometry.helpers_materials import *
from .source.helpers_source import *
from .actor.helpers_actor import *
from .actor.helpers_filter import *
from .SimulationUserInfo import *
from .helpers_element import *

# Volume specific
from .geometry.MaterialBuilder import *
from .geometry.MaterialDatabase import *
from .geometry.VolumeManager import *
from .geometry.SolidBuilderBase import *

# Source specific
from .source.SourceManager import *
from .source.GANSourceConditionalGenerator import *
from .source.GANSourceConditionalPairsGenerator import *
from .source.VoxelizedSourceConditionGenerator import *
from .helpers_physics import *

# Actor specific
from .actor.FilterManager import *
from .actor.ActorManager import *
from .ActionManager import *
from .UIsessionSilent import *
from .UIsessionVerbose import *
from .RunAction import *

# Physics
from .physics.PhysicsUserInfo import *
from .physics.PhysicsManager import *



from .VoxelsSource import *
from .GANSource import *
from .PBSource import *
import pathlib

"""
    List of source types: Generic, Voxels etc

    Energy spectra for beta+ emitters
"""

source_type_names = {GenericSource, VoxelsSource, GANSource, PBSource}
source_builders = gate.make_builders(source_type_names)

gate_source_path = pathlib.Path(__file__).parent.resolve()

# http://www.lnhb.fr/nuclear-data/module-lara/
all_beta_plus_radionuclides = [
    "F18",
    "Ga68",
    "Zr89",
    "Na22",
    "C11",
    "N13",
    "O15",
    "Rb82",
]


def read_beta_plus_spectra(rad_name):
    """
    read the file downloaded from LNHB
    there are 15 lines-long header to skip
    first column is E(keV)
    second column is dNtot/dE b+
    WARNING : bins width is not uniform (need to scale for density)
    """
    filename = (
        f"{gate_source_path}/beta_plus_spectra/{rad_name}/beta+_{rad_name}_tot.bs"
    )
    data = np.genfromtxt(filename, usecols=(0, 1), skip_header=15, dtype=float)
    return data


def compute_bins_density(bins):
    """
    Given a list of (energy) bins center, compute the width of each bin.
    """
    lower = np.roll(bins, 1)
    lower[0] = 0
    upper = bins
    dx = upper - lower
    return dx


def get_rad_yield(rad_name):
    if not rad_name in all_beta_plus_radionuclides:
        return 1.0
    data = read_beta_plus_spectra(rad_name)
    ene = data[:, 0] / 1000  # convert from KeV to MeV
    proba = data[:, 1]
    cdf, total = gate.compute_cdf_and_total_yield(proba, ene)
    total = total * 1000  # (because was in MeV)
    return total


def compute_cdf_and_total_yield(data, bins):
    """
    Compute the CDF (Cumulative Density Function) of a list of non-uniform energy bins
    with associated probability.
    Also return the total probability.
    """
    dx = compute_bins_density(bins)
    p = data * dx
    total = p.sum()
    cdf = np.cumsum(p) / total
    return cdf, total


def generate_isotropic_directions(
    n, min_theta=0, max_theta=np.pi, min_phi=0, max_phi=2 * np.pi, rs=np.random
):
    """
    like in G4SPSAngDistribution.cc

    Later : do a version with torch (gpu) instead of np (cpu) ?
    """
    u = rs.uniform(0, 1, size=n)
    costheta = np.cos(min_theta) - u * (np.cos(min_theta) - np.cos(max_theta))
    sintheta = np.sqrt(1 - costheta**2)

    v = rs.uniform(0, 1, size=n)
    phi = min_phi + (max_phi - min_phi) * v
    sinphi = np.sin(phi)
    cosphi = np.cos(phi)

    # "direct cosine" method, like in Geant4 (already normalized)
    px = -sintheta * cosphi
    py = -sintheta * sinphi
    pz = -costheta

    # concat
    v = np.column_stack((px, py, pz))

    return v


def get_rad_energy_spectrum(rad):
    weights = {}
    energies = {}
    MeV = gate.g4_units("MeV")
    # Tc99m
    weights["Tc99m"] = [0.885]
    energies["Tc99m"] = [0.140511 * MeV]
    # Lu177
    weights["Lu177"] = [0.001726, 0.0620, 0.000470, 0.1038, 0.002012, 0.00216]
    energies["Lu177"] = [
        0.0716418 * MeV,
        0.1129498 * MeV,
        0.1367245 * MeV,
        0.2083662 * MeV,
        0.2496742 * MeV,
        0.3213159 * MeV,
    ]

    # In111
    weights["In111"] = [0.000015, 0.9061, 0.9412]
    energies["In111"] = [0.15081 * MeV, 0.17128 * MeV, 0.24535 * MeV]
    # I131
    weights["I131"] = [
        0.02607,
        0.000051,
        0.000211,
        0.00277,
        0.000023,
        0.000581,
        0.0614,
        0.000012,
        0.000046,
        0.000807,
        0.000244,
        0.00274,
        0.00017,
        0.812,
        0.000552,
        0.003540,
        0.0712,
        0.002183,
        0.01786,
    ]
    energies["I131"] = [
        0.080185 * MeV,
        0.0859 * MeV,
        0.163930 * MeV,
        0.177214 * MeV,
        0.23218 * MeV,
        0.272498 * MeV,
        0.284305 * MeV,
        0.2958 * MeV,
        0.3024 * MeV,
        0.318088 * MeV,
        0.324651 * MeV,
        0.325789 * MeV,
        0.3584 * MeV,
        0.364489 * MeV,
        0.404814 * MeV,
        0.503004 * MeV,
        0.636989 * MeV,
        0.642719 * MeV,
        0.722911 * MeV,
    ]

    return weights[rad], energies[rad]


def set_source_rad_energy_spectrum(source, rad):
    w, en = get_rad_energy_spectrum(rad)
    source.energy.type = "spectrum"
    source.energy.spectrum_weight = w
    source.energy.spectrum_energy = en



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1
ui.check_volumes_overlap = False

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq

# world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]
world.material = "G4_AIR"

# spect head (debug mode = very small collimator)
spect = gate_spect.add_ge_nm67_spect_head(sim, "spect", debug=False)
psd = 6.11 * cm
spect.translation = [0, 0, -(20 * cm + psd)]

# waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [15 * cm, 15 * cm, 15 * cm]
waterbox.material = "G4_WATER"
blue = [0, 1, 1, 1]
waterbox.color = blue

# physic list
p = sim.get_physics_user_info()
p.physics_list_name = "G4EmStandardPhysics_option4"
p.enable_decay = False
cuts = p.production_cuts
cuts.world.gamma = 10 * mm
cuts.world.electron = 10 * mm
cuts.world.positron = 10 * mm
cuts.world.proton = 10 * mm

cuts.spect.gamma = 0.1 * mm
cuts.spect.electron = 0.1 * mm
cuts.spect.positron = 0.1 * mm

# default source for tests
activity = 100 * kBq
beam1 = sim.add_source("Generic", "beam1")
beam1.mother = waterbox.name
beam1.particle = "gamma"
beam1.energy.mono = 140.5 * keV
beam1.position.type = "sphere"
beam1.position.radius = 3 * cm
beam1.position.translation = [0, 0, 0 * cm]
beam1.direction.type = "momentum"
beam1.direction.momentum = [0, 0, -1]
beam1.activity = activity / ui.number_of_threads

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.track_types_flag = True

# create G4 objects
sim.initialize()

# start simulation
sim.start()

# stat
gate.warning("Compare stats")
stats = sim.get_actor("Stats")
print(stats)
print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
stats.counts.run_count = 1  # force to 1
stats_ref = gate.read_stat_file(paths.gate_output / "stat1.txt")
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.02)

gate.test_ok(is_ok)



import colored
import numpy as np
import opengate as gate
import opengate_core as g4
from box import Box
from anytree import RenderTree
import textwrap
from inspect import getframeinfo, stack
import pkg_resources
import sys
from pathlib import Path
import random
import string
import os
from numpy.random import MT19937
from numpy.random import RandomState, SeedSequence
import inspect
import re
import json

color_error = colored.fg("red") + colored.attr("bold")
color_warning = colored.fg("orange_1")
color_ok = colored.fg("green")


def fatal(s):
    caller = getframeinfo(stack()[1][0])
    ss = f"(in {caller.filename} line {caller.lineno})"
    ss = colored.stylize(ss, color_error)
    gate.log.critical(ss)
    s = colored.stylize(s, color_error)
    gate.log.critical(s)
    sys.exit(-1)


def warning(s):
    s = colored.stylize(s, color_warning)
    gate.log.warning(s)


def raise_except(s):
    s = colored.stylize(s, color_error)
    raise Exception(s)


def pretty_print_tree(tree, geometry):
    """Print tree"""
    s = ""
    for pre, fill, node in RenderTree(tree[gate.__world_name__]):
        v = geometry[node.name]
        s += f"{pre}{node.name} {v.type_name} {v.material}\n"

    # remove last break line
    return s[:-1]


def assert_equal_dic(d1, d2, name=""):
    for k in d1:
        if not k in d2:
            fatal(f"ERROR missing key {k} in {name}")
        if isinstance(d1[k], np.ndarray):
            if np.any(d2[k] != d1[k]):
                fatal(f"ERROR np array {k} {d1[k]} in {name}")
        else:
            if d2[k] != d1[k]:
                fatal(f"ERROR value for {k} in {name}")
    for k in d2:
        if not k in d1:
            fatal(f"ERROR, additional key {k} in {name}")


def g4_units(name: str) -> float:
    table = g4.G4UnitDefinition.GetUnitsTable()
    for t in table:
        for a in t.GetUnitsList():
            if a.GetName() == name or a.GetSymbol() == name:
                return a.GetValue()
    units_list = []
    for t in table:
        for a in t.GetUnitsList():
            units_list.append(a.GetSymbol())
    s = [str(u) + " " for u in units_list]
    fatal(f"Error, cannot find the unit named {name}. Known are: {s}")


def g4_best_unit(value, unit_type):
    return g4.G4BestUnit(value, unit_type)


def assert_key(key: str, d: Box):
    if key not in d:
        gate.fatal(f'The key "{key}" is needed in this structure:\n' f"{d}")


def assert_keys(keys: list, d: Box):
    for key in keys:
        assert_key(key, d)


def indent(amount, text, ch=" "):
    """
    Prefix the text with indent spaces
    https://stackoverflow.com/questions/8234274/how-to-indent-the-contents-of-a-multi-line-string
    """
    return textwrap.indent(text, amount * ch)


def assert_unique_element_name(elements, name):
    if name in elements:
        s = (
            f"Error, cannot add '{name}' because this element's name already exists"
            f" in: {elements}."
        )
        gate.fatal(s)


def make_builders(class_names):
    """
    Consider a list of Classname. For each, it build a key/value, with:
    - the type of the class as key
    - and a lambda function that create an object of this class as value
    """
    builder_list = {}
    for c in class_names:
        # note the following lambda:
        # https://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture
        try:
            builder_list[c.type_name] = lambda x, y=c: y(x)
        except AttributeError:
            # if type_name is not an attribute of the class,
            # we use the name of the class as key.
            # Also: no name parameter (this is for Physics List)
            builder_list[c.__name__] = lambda y=c: y()
    return builder_list


def read_mac_file_to_commands(filename):
    # read a file located into the 'mac' folder of the source code
    # return a list of commands
    resource_package = __name__
    resource_path = "/".join(("mac", filename))  # Do not use os.path.join()
    template = pkg_resources.resource_string(resource_package, resource_path)
    c = template.decode("utf-8")
    commands = []
    for s in c.split("\n"):
        if s == "":
            continue
        # if s[0] == '#':
        #    continue
        commands.append(s)
    return commands


def check_filename_type(filename):
    # Algorithms (itk) do not support Path -> convert to str
    if isinstance(filename, Path):
        return str(filename)
    return filename


def get_random_folder_name(size=8, create=True):
    r = "".join(random.choices(string.ascii_lowercase + string.digits, k=size))
    r = "run." + r
    if create:
        if not os.path.exists(r):
            print(f"Creating output folder {r}")
            os.mkdir(r)
        if not os.path.isdir(r):
            gate.fatal(f"Error, while creating {r}.")
    return r


def import_gaga_phsp():
    # Try to import torch
    try:
        import torch
    except:
        gate.fatal(
            f'The module "torch" is needed, see https://pytorch.org/get-started/locally/ to install it'
        )

    # Try to import gaga_phsp
    try:
        import gaga_phsp as gaga
    except:
        gate.fatal("The module \"gaga_phsp\" is needed. Use ' pip install gaga_phsp'")

    # Check minimal version of gaga_phsp
    import pkg_resources
    from packaging import version

    gaga_version = pkg_resources.get_distribution("gaga_phsp").version
    gaga_minimal_version = "0.5.8"
    if version.parse(gaga_version) < version.parse(gaga_minimal_version):
        gate.fatal(
            "The minimal version of gaga_phsp is not correct. You should install at least the version "
            + gaga_minimal_version
            + ". Your version is "
            + gaga_version
        )
    return gaga


def import_garf():
    # Try to import torch
    try:
        import torch
    except:
        gate.fatal(
            f'The module "torch" is needed, see https://pytorch.org/get-started/locally/ to install it'
        )

    # Try to import garf_phsp
    try:
        import garf
    except:
        gate.fatal("The module \"garf\" is needed. Use ' pip install garf'")

    # Check minimal version of garf
    import pkg_resources
    from packaging import version

    garf_version = pkg_resources.get_distribution("garf").version
    garf_minimal_version = "2.2"
    if version.parse(garf_version) < version.parse(garf_minimal_version):
        gate.fatal(
            "The minimal version of garf is not correct. You should install at least the version "
            + garf_minimal_version
            + ". Your version is "
            + garf_version
        )
    return garf


def get_rnd_seed(seed):
    return RandomState(MT19937(SeedSequence(seed)))


def DDF():
    """
    Debug print current Function name
    """
    print("--> Entering", inspect.stack()[1][3])


def DD(arg):
    """
    Debug print variable name and its value
    """
    frame = inspect.currentframe()
    try:
        context = inspect.getframeinfo(frame.f_back).code_context
        caller_lines = "".join([line.strip() for line in context])
        m = re.search(r"DD\s*\((.+?)\);*$", caller_lines)
        if m:
            caller_lines = m.group(1)
            # end if
        print(caller_lines, "=", arg)
    finally:
        del frame


def print_dic(dic):
    print(json.dumps(dic, indent=4, default=str))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test010_generic_source")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.number_of_threads = 1

# set the world size like in the Gate macro
m = gate.g4_units("m")
world = sim.world
world.size = [2 * m, 2 * m, 2 * m]

# add a simple volume
waterbox = sim.add_volume("Box", "waterbox")
cm = gate.g4_units("cm")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
waterbox.material = "G4_WATER"

# useful units
MeV = gate.g4_units("MeV")
keV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
deg = gate.g4_units("deg")
mm = gate.g4_units("mm")

# test sources
source = sim.add_source("Generic", "source1")
source.particle = "gamma"
source.activity = 10000 * Bq / ui.number_of_threads
source.position.type = "sphere"
source.position.radius = 5 * mm
source.position.translation = [-3 * cm, 30 * cm, -3 * cm]
source.direction.type = "momentum"
source.direction.momentum = [0, -1, 0]
source.energy.type = "mono"
source.energy.mono = 1 * MeV

source = sim.add_source("Generic", "source2")
source.particle = "proton"
source.activity = 10000 * Bq / ui.number_of_threads
source.position.type = "disc"
source.position.radius = 5 * mm
source.position.translation = [6 * cm, 5 * cm, -30 * cm]
# source.position.rotation = Rotation.from_euler('x', 45, degrees=True).as_matrix()
source.position.rotation = Rotation.identity().as_matrix()
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.energy.type = "gauss"
source.energy.mono = 140 * MeV
source.energy.sigma_gauss = 10 * MeV

source = sim.add_source("Generic", "s3")
source.particle = "proton"
source.activity = 10000 * Bq / ui.number_of_threads
source.position.type = "box"
source.position.size = [4 * cm, 4 * cm, 4 * cm]
source.position.translation = [8 * cm, 8 * cm, 30 * cm]
source.direction.type = "focused"
source.direction.focus_point = [1 * cm, 2 * cm, 3 * cm]
source.energy.type = "gauss"
source.energy.mono = 140 * MeV
source.energy.sigma_gauss = 10 * MeV

source = sim.add_source("Generic", "s4")
source.particle = "proton"
source.activity = 10000 * Bq / ui.number_of_threads
source.position.type = "box"
source.position.size = [4 * cm, 4 * cm, 4 * cm]
source.position.translation = [-3 * cm, -3 * cm, -3 * cm]
# source.position.rotation = Rotation.from_euler('x', 45, degrees=True).as_matrix()
source.position.rotation = Rotation.identity().as_matrix()
source.direction.type = "iso"
source.energy.type = "gauss"
source.energy.mono = 80 * MeV
source.energy.sigma_gauss = 1 * MeV

# actors
stats = sim.add_actor("SimulationStatisticsActor", "Stats")

# src_info = sim.add_actor('SourceInfoActor', 'src_info')
# src_info.filename = 'output/sources.root'

dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test010-edep.mhd"
dose.mother = "waterbox"
dose.size = [50, 50, 50]
dose.spacing = [4 * mm, 4 * mm, 4 * mm]

# create G4 objects
sim.initialize()

# print after init
print(sim)
print("Simulation seed:", sim.actual_random_seed)

# verbose
sim.apply_g4_command("/tracking/verbose 0")
# sim.apply_g4_command("/run/verbose 2")
# sim.apply_g4_command("/event/verbose 2")
# sim.apply_g4_command("/tracking/verbose 1")

# start simulation

sim.start()

# get results
stats = sim.get_actor("Stats")
print(stats)

dose = sim.get_actor("dose")
print(dose)

# gate_test10
# Gate mac/main.mac
# Current version is two times slower :(
stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
print("-" * 80)
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.05)
is_ok = is_ok and gate.assert_images(
    paths.gate_output / "output-Edep.mhd",
    paths.output / "test010-edep.mhd",
    stats,
    tolerance=30,
)

gate.test_ok(is_ok)



import opengate as gate
import pathlib

# unit
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
deg = gate.g4_units("deg")

# colors
red = [1, 0.7, 0.7, 0.8]
blue = [0.5, 0.5, 1, 0.8]
gray = [0.5, 0.5, 0.5, 1]
white = [1, 1, 1, 1]
yellow = [1, 1, 0, 1]
green = [0, 1, 0, 1]


def get_collimator(rad):
    radionuclides = ["Tc99m", "Lu177", "In111", "I131"]
    ref_collimators = ["lehr", "megp", "megp", "hegp"]
    if rad not in radionuclides:
        gate.fatal(f'The radionuclide "{rad}" is unknown. Known are: {radionuclides}')
    return ref_collimators[radionuclides.index(rad)]


def add_ge_nm67_fake_spect_head(sim, name="spect"):
    spect_length = 19 * cm
    head = sim.add_volume("Box", name)
    head.material = "G4_AIR"
    head.size = [57.6 * cm, 44.6 * cm, spect_length]
    head.color = white
    return head


def get_orientation_for_CT(colli_type, table_shift, radius):
    nm = gate.g4_units("nm")
    pos, crystal_distance, psdd = get_plane_position_and_distance_to_crystal(colli_type)
    pos += 1 * nm
    p = [0, table_shift, -(radius + psdd)]
    return gate.get_transform_orbiting(p, "x", 90)


def add_ge_nm67_spect_head(sim, name="spect", collimator_type="lehr", debug=False):
    """
    Collimators:
    - False : no collimator
    - lehr : holes length 35 mm, diam 1.5 mm, septal thickness : 0.2 mm
    - megp : holes length 58 mm, diam 3 mm,   septal thickness : 1.05 mm
    - hegp : holes length 66 mm, diam 4 mm,   septal thickness : 1.8 mm

    Collimator LEHR: Low Energy High Resolution    (for Tc99m)
    Collimator MEGP: Medium Energy General Purpose (for In111, Lu177)
    Collimator HEGP: High Energy General Purpose   (for I131)

    """
    f = pathlib.Path(__file__).parent.resolve()
    fdb = f"{f}/spect_ge_nm670_materials.db"
    if fdb not in sim.volume_manager.material_databases:
        sim.add_material_database(fdb)

    # check overlap
    sim.g4_check_overlap_flag = False  # set to True for debug

    # spect head
    head, lead_cover = add_ge_nm670_spect_box(sim, name, collimator_type)

    # spect head
    crystal = add_ge_nm670_spect_crystal(sim, name, lead_cover)

    # spect collimator
    if collimator_type:
        colli = add_ge_nm670_spect_collimator(sim, name, head, collimator_type, debug)

    return head


def distance_to_center_of_crystal(sim, name="spect"):
    lead_cover = sim.get_volume_user_info(f"{name}_lead_cover")
    crystal = sim.get_volume_user_info(f"{name}_crystal")
    # distance from center to center of crystal
    shielding = sim.get_volume_user_info(f"{name}_shielding")
    d = shielding.translation[2] + lead_cover.translation[2] + crystal.translation[2]
    return d


def add_ge_nm670_spect_box(sim, name, collimator_type):
    # the total length
    spect_length = 19 * cm

    # bounding box
    head = sim.add_volume("Box", name)
    head.material = "G4_AIR"
    head.size = [57.6 * cm, 44.6 * cm, spect_length]
    head.color = white

    # shielding
    shielding = sim.add_volume("Box", f"{name}_shielding")
    shielding.mother = head.name
    shielding.size = head.size.copy()
    shielding.size[2] = 11.1375 * cm
    shielding.translation = [0, 0, -3.64 * cm]
    shielding.material = "Steel"
    shielding.color = yellow

    # shielding lead cover
    lead_cover = sim.add_volume("Box", f"{name}_lead_cover")
    lead_cover.mother = shielding.name
    lead_cover.size = [57.6 * cm, 40.6 * cm, 10.1375 * cm]
    lead_cover.translation = [0, 0, 0.5 * cm]
    lead_cover.material = "Lead"
    lead_cover.color = gray

    # shielding alu cover
    alu_cover = sim.add_volume("Box", f"{name}_alu_cover")
    alu_cover.mother = lead_cover.name
    alu_cover.size = [54 * cm, 40 * cm, 0.13 * cm]
    alu_cover.translation = [0, 0, 5.00375 * cm]
    alu_cover.material = "Aluminium"
    alu_cover.color = blue

    # shielding reflector
    reflector = sim.add_volume("Box", f"{name}_reflector")
    reflector.mother = lead_cover.name
    reflector.size = [54 * cm, 40 * cm, 0.12 * cm]
    reflector.translation = [0, 0, 3.92625 * cm]
    reflector.material = "TiO2"
    reflector.color = green

    # backside
    # The back-side is fairly complex, and may have a strong influence on the
    # spectrum: the model shown here is simplified
    backside = sim.add_volume("Box", f"{name}_backside")
    backside.mother = lead_cover.name
    backside.size = [54 * cm, 40 * cm, 8 * cm]
    backside.translation = [0, 0, -0.13375 * cm]
    backside.material = "Pyrex66"
    backside.color = blue

    return head, lead_cover


def add_ge_nm670_spect_crystal(sim, name, lead_cover):
    # mono-bloc crystal thickness 3/8 of inch = 0.9525 cm
    # (if 5/8 inch = 1.5875 ; but probably need to translate elements)
    crystal = sim.add_volume("Box", f"{name}_crystal")
    crystal.mother = lead_cover.name
    crystal.size = [54 * cm, 40 * cm, 0.9525 * cm]
    crystal.translation = [0, 0, 4.4625 * cm]
    crystal.material = "NaITl"
    crystal.color = yellow
    return crystal


def add_ge_nm670_spect_collimator(sim, name, head, collimator_type, debug):
    """
    Start with default lehr collimator description,
    then change some parameters for the other types
    """

    # mono-bloc crystal thickness 3/8 of inch
    colli_trd = sim.add_volume("Trd", f"{name}_collimator_trd")
    colli_trd.mother = head.name
    colli_trd.dx2 = 56.8 * cm / 2.0
    colli_trd.dy2 = 42.8 * cm / 2.0
    colli_trd.dx1 = 57.6 * cm / 2.0
    colli_trd.dy1 = 44.6 * cm / 2.0
    colli_trd.dz = 4.18 * cm / 2.0  # dep. on colli type
    colli_trd.translation = [0, 0, 4.02 * cm]  # dep. on colli type
    colli_trd.material = "G4_AIR"
    colli_trd.color = red

    # PSD (Position Sensitive Detection)
    psd = sim.add_volume("Box", f"{name}_collimator_psd")
    psd.mother = colli_trd.name
    psd.size = [54.6 * cm, 40.6 * cm, 0.1 * cm]
    psd.translation = [0, 0, 2.04 * cm]  # dep. on colli type
    psd.material = "Aluminium"
    psd.color = green

    # PSD layer
    psd_layer = sim.add_volume("Box", f"{name}_collimator_psd_layer")
    psd_layer.mother = colli_trd.name
    psd_layer.size = [54.6 * cm, 40.6 * cm, 0.15 * cm]
    psd_layer.translation = [0, 0, 1.915 * cm]  # dep. on colli type
    psd_layer.material = "PVC"
    psd_layer.color = red

    # Alu cover
    alu_cover = sim.add_volume("Box", f"{name}_collimator_alu_cover")
    alu_cover.mother = colli_trd.name
    alu_cover.size = [54.6 * cm, 40.6 * cm, 0.05 * cm]
    alu_cover.translation = [0, 0, -2.065 * cm]  # dep. on colli type
    alu_cover.material = "Aluminium"
    alu_cover.color = blue

    # air gap
    air_gap = sim.add_volume("Box", f"{name}_collimator_air_gap")
    air_gap.mother = colli_trd.name
    air_gap.size = [54.6 * cm, 40.6 * cm, 0.38 * cm]
    air_gap.translation = [0, 0, 1.65 * cm]  # dep. on colli type
    air_gap.material = "G4_AIR"
    air_gap.color = blue

    # core
    core = sim.add_volume("Box", f"{name}_collimator_core")
    core.mother = colli_trd.name
    core.size = [54.6 * cm, 40.6 * cm, 3.5 * cm]  # dep. on colli type
    core.translation = [0, 0, -0.29 * cm]
    core.material = "Lead"
    core.color = blue

    # adapt according to collimator type
    if collimator_type == "megp":
        colli_trd.dz = 6.48 * cm / 2.0
        colli_trd.translation = [0, 0, 5.17 * cm]
        psd.translation = [0, 0, 3.19 * cm]
        psd_layer.translation = [0, 0, 3.065 * cm]
        alu_cover.translation = [0, 0, -3.215 * cm]
        air_gap.translation = [0, 0, 2.80 * cm]
        core.size = [54.6 * cm, 40.6 * cm, 5.8 * cm]

    if collimator_type == "hegp":
        colli_trd.dz = 7.28 * cm / 2.0
        colli_trd.translation = [0, 0, 5.57 * cm]
        psd.translation = [0, 0, 3.59 * cm]
        psd_layer.translation = [0, 0, 3.465 * cm]
        alu_cover.translation = [0, 0, -3.615 * cm]
        air_gap.translation = [0, 0, 3.2 * cm]
        core.size = [54.6 * cm, 40.6 * cm, 6.6 * cm]

    # repeater for the holes
    holep = False
    if collimator_type == "megp":
        holep = megp_collimator_repeater(sim, name, core, debug)
    if collimator_type == "lehr":
        holep = lehr_collimator_repeater(sim, name, core, debug)
    if collimator_type == "hegp":
        holep = hegp_collimator_repeater(sim, name, core, debug)
    if not holep:
        gate.fatal(
            f"Error, unknown collimator type {collimator_type}. "
            f'Use "megp" or "lehr" or "hegp" or "False"'
        )

    return colli_trd


def hegp_collimator_repeater(sim, name, core, debug):
    # one single hole
    hole = sim.add_volume("Polyhedra", f"{name}_collimator_hole")
    hole.phi_start = 0 * deg
    hole.phi_total = 360 * deg
    hole.num_side = 6
    hole.num_zplanes = 2
    h = 6.6 * cm
    hole.zplane = [-h / 2, h - h / 2]
    hole.radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    r = 0.2 * cm
    hole.radius_outer = [r] * hole.num_side
    hole.material = "G4_AIR"
    hole.mother = core.name

    # parameterised holes
    size = [54, 70, 1]
    if debug:
        size = [10, 10, 1]
    tr = [10.0459 * mm, 5.8 * mm, 0]
    holep = gate.build_param_repeater(sim, core.name, hole.name, size, tr)

    # dot it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = [5.0229 * mm, 2.9000 * mm, 0]

    return holep


def megp_collimator_repeater(sim, name, core, debug):
    # one single hole
    hole = sim.add_volume("Polyhedra", f"{name}_collimator_hole")
    hole.phi_start = 0 * deg
    hole.phi_total = 360 * deg
    hole.num_side = 6
    hole.num_zplanes = 2
    h = 5.8 * cm
    hole.zplane = [-h / 2, h - h / 2]
    hole.radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    r = 0.15 * cm
    hole.radius_outer = [r] * hole.num_side
    hole.material = "G4_AIR"
    hole.mother = core.name

    # parameterised holes
    size = [77, 100, 1]
    if debug:
        size = [10, 10, 1]
    tr = [7.01481 * mm, 4.05 * mm, 0]
    holep = gate.build_param_repeater(sim, core.name, hole.name, size, tr)

    # dot it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = [3.50704 * mm, 2.025 * mm, 0]

    return holep


def lehr_collimator_repeater(sim, name, core, debug):
    # one single hole
    hole = sim.add_volume("Polyhedra", f"{name}_collimator_hole")
    hole.phi_start = 0 * deg
    hole.phi_total = 360 * deg
    hole.num_side = 6
    hole.num_zplanes = 2
    h = 3.5 * cm
    hole.zplane = [-h / 2, h - h / 2]
    hole.radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    r = 0.075 * cm
    hole.radius_outer = [r] * hole.num_side
    hole.material = "G4_AIR"
    hole.mother = core.name

    # parameterised holes
    size = [183, 235, 1]
    if debug:
        size = [10, 10, 1]
    tr = [2.94449 * mm, 1.7 * mm, 0]
    holep = gate.build_param_repeater(sim, core.name, hole.name, size, tr)

    # dot it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = [1.47224 * mm, 0.85 * mm, 0]

    return holep


def UNUSED_megp_collimator_repeater_parametrised(sim, name, core, debug):
    """# because this volume will be parameterised, we need to prevent
    # the creation of the physical volume
    hole.build_physical_volume = False

    # parameterised holes
    holep = sim.add_volume('RepeatParametrised', f'{name}_collimator_hole_param')
    holep.mother = core.name
    holep.translation = None
    holep.rotation = None
    holep.repeated_volume_name = hole.name
    # number of repetition
    holep.linear_repeat = [183, 235, 1]
    if debug:
        holep.linear_repeat = [10, 10, 1]
    # translation for each repetition
    holep.translation = [2.94449 * mm, 1.7 * mm, 0]
    # starting position
    holep.start = [-(holep.linear_repeat[0] * holep.translation[0]) / 2.0,
                   -(holep.linear_repeat[1] * holep.translation[1]) / 2.0, 0]
    """


def add_simplified_digitizer_Tc99m(
    sim, crystal_volume_name, output_name, scatter_flag=False
):
    # units
    keV = gate.g4_units("keV")
    # default  channels
    channels = []
    if scatter_flag:
        channels = [
            {
                "name": f"scatter_{crystal_volume_name}",
                "min": 114 * keV,
                "max": 126 * keV,
            }
        ]
    channels.append(
        {
            "name": f"peak140_{crystal_volume_name}",
            "min": 126 * keV,
            "max": 154.55 * keV,
        }
    )
    proj = add_digitizer(sim, crystal_volume_name, channels)
    # output
    proj.output = output_name
    return proj


def get_simplified_digitizer_channels_rad(spect_name, rad, scatter_flag):
    all_channels = {}
    keV = gate.g4_units("keV")

    # Tc99m
    channels = [
        {"name": f"scatter_{spect_name}", "min": 114 * keV, "max": 126 * keV},
        {"name": f"peak140_{spect_name}", "min": 126 * keV, "max": 154 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
    all_channels["Tc99m"] = channels

    # Lu177, Ljungberg2016
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 96 * keV, "max": 104 * keV},
        {"name": f"peak113_{spect_name}", "min": 104 * keV, "max": 121.48 * keV},
        {"name": f"scatter2_{spect_name}", "min": 122.48 * keV, "max": 133.12 * keV},
        {"name": f"scatter3_{spect_name}", "min": 176.46 * keV, "max": 191.36 * keV},
        {"name": f"peak208_{spect_name}", "min": 192.36 * keV, "max": 223.6 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
    all_channels["Lu177"] = channels

    # In111
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 150 * keV, "max": 156 * keV},
        {"name": f"peak171_{spect_name}", "min": 156 * keV, "max": 186 * keV},
        {"name": f"scatter2_{spect_name}", "min": 186 * keV, "max": 192 * keV},
        {"name": f"scatter3_{spect_name}", "min": 218 * keV, "max": 224 * keV},
        {"name": f"peak245_{spect_name}", "min": 224 * keV, "max": 272 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
    all_channels["In111"] = channels

    # I131
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 314 * keV, "max": 336 * keV},
        {"name": f"peak364_{spect_name}", "min": 336 * keV, "max": 392 * keV},
        {"name": f"scatter2_{spect_name}", "min": 392 * keV, "max": 414 * keV},
        {"name": f"scatter3_{spect_name}", "min": 414 * keV, "max": 556 * keV},
        {"name": f"scatter4_{spect_name}", "min": 556 * keV, "max": 595 * keV},
        {"name": f"peak637_{spect_name}", "min": 595 * keV, "max": 679 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
        channels.pop(1)
    all_channels["I131"] = channels

    # choose
    channels = all_channels[rad]
    return channels


def add_digitizer(sim, crystal_volume_name, channels):
    # units
    mm = gate.g4_units("mm")
    cc = add_digitizer_energy_windows(sim, crystal_volume_name, channels)

    # projection
    proj = sim.add_actor("HitsProjectionActor", f"Projection_{crystal_volume_name}")
    proj.mother = cc.mother
    proj.input_hits_collections = [x["name"] for x in cc.channels]
    # proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.spacing = [5 * mm, 5 * mm]
    proj.size = [128, 128]

    return proj


def add_digitizer_energy_windows(sim, crystal_volume_name, channels):
    hc = sim.add_actor("HitsCollectionActor", f"Hits_{crystal_volume_name}")
    hc.mother = crystal_volume_name
    hc.output = ""  # No output
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PostStepUniqueVolumeID",
        "GlobalTime",
    ]
    sc = sim.add_actor("HitsAdderActor", f"Singles_{crystal_volume_name}")
    sc.mother = hc.mother
    sc.input_hits_collection = hc.name
    sc.policy = "EnergyWinnerPosition"
    sc.output = ""  # No output
    cc = sim.add_actor("HitsEnergyWindowsActor", f"EnergyWindows_{crystal_volume_name}")
    cc.mother = sc.mother
    cc.input_hits_collection = sc.name
    cc.channels = channels
    cc.output = ""  # No output
    return cc


def get_volume_position_in_head(sim, spect_name, vol_name, pos="max"):
    vol = sim.get_volume_user_info(f"{spect_name}_{vol_name}")
    pMin, pMax = gate.get_volume_bounding_limits(sim, vol.name)
    x = pMax
    if pos == "min":
        x = pMin
    if pos == "center":
        x = pMin + (pMax - pMin) / 2.0
    x = gate.vec_g4_as_np(x)
    x = gate.translate_point_to_volume(sim, vol, spect_name, x)
    return x[2]


def get_plane_position_and_distance_to_crystal(collimator_type):
    """
    This has been computed with t043_distances
    - first : distance from head center to the PSD (translation for the plane)
    - second: distance from PSD to center of the crystal
    - third : distance from the head boundary to the PSD (for spect_radius info)
    """
    if collimator_type == "lehr":
        return 61.1, 47.875, 33.9

    if collimator_type == "megp":
        return 84.1, 70.875, 10.9

    if collimator_type == "hegp":
        return 92.1, 78.875, 2.9

    gate.fatal(
        f'Unknown collimator type "{collimator_type}", please use lehr or megp or hegp'
    )



import opengate as gate
import opengate_core as g4
from anytree import LevelOrderIter
import numpy as np
import math
import time

iec_plastic = "IEC_PLASTIC"
water = "G4_WATER"
iec_lung = "G4_LUNG_ICRP"


def create_material():
    n = g4.G4NistManager.Instance()
    elems = ["C", "H", "O"]
    nbAtoms = [5, 8, 2]
    gcm3 = gate.g4_units("g/cm3")
    n.ConstructNewMaterialNbAtoms("IEC_PLASTIC", elems, nbAtoms, 1.18 * gcm3)


def add_phantom(simulation, name="iec"):
    # unit
    cm = gate.g4_units("cm")
    mm = gate.g4_units("mm")
    deg = gate.g4_units("deg")
    create_material()

    # colors
    red = [1, 0.7, 0.7, 0.8]
    blue = [0.5, 0.5, 1, 0.8]
    gray = [0.5, 0.5, 0.5, 1]

    # check overlap
    simulation.g4_check_overlap_flag = False

    # top
    top_shell = simulation.new_solid("Tubs", f"{name}_top_shell")
    top_shell.rmax = 15 * cm
    top_shell.rmin = 0
    top_shell.dz = 21.4 * cm / 2
    top_shell.sphi = 0 * deg
    top_shell.dphi = 180 * deg

    # Lower left half of phantom
    bottom_left_shell = simulation.new_solid("Tubs", f"{name}_bottom_left_shell")
    bottom_left_shell.rmax = 8 * cm
    bottom_left_shell.rmin = 0
    bottom_left_shell.dz = 21.4 * cm / 2
    bottom_left_shell.sphi = 270 * deg
    bottom_left_shell.dphi = 90 * deg

    # Lower right half of phantom
    bottom_right_shell = simulation.new_solid("Tubs", f"{name}_bottom_right_shell")
    gate.copy_user_info(bottom_left_shell, bottom_right_shell)
    bottom_right_shell.sphi = 180 * deg

    # Bottom box
    bottom_central_shell = simulation.new_solid("Box", f"{name}_bottom_central_shell")
    bottom_central_shell.size = [14 * cm, 8 * cm, 21.4 * cm]

    c = -bottom_central_shell.size[1] / 2

    # union
    shell = gate.solid_union(top_shell, bottom_left_shell, [7 * cm, 0, 0])
    shell = gate.solid_union(shell, bottom_central_shell, [0, c, 0])
    shell = gate.solid_union(shell, bottom_right_shell, [-7 * cm, 0, 0])
    iec = simulation.add_volume_from_solid(shell, name)
    iec.material = iec_plastic
    iec.color = red

    # Inside space for the water, same than the shell, with 0.3cm less
    thickness = 0.3 * cm
    top_interior = simulation.new_solid("Tubs", f"{name}_top_interior")
    gate.copy_user_info(top_shell, top_interior)
    top_interior.rmax -= thickness
    top_interior.dz -= thickness
    bottom_left_interior = simulation.new_solid("Tubs", f"{name}_bottom_left_interior")
    gate.copy_user_info(bottom_left_shell, bottom_left_interior)

    bottom_left_interior.rmax -= thickness
    bottom_left_interior.dz -= thickness
    bottom_right_interior = simulation.new_solid(
        "Tubs", f"{name}_bottom_right_interior"
    )
    gate.copy_user_info(bottom_left_interior, bottom_right_interior)
    bottom_right_interior.sphi = 180 * deg
    bottom_central_interior = simulation.new_solid(
        "Box", f"{name}_bottom_central_interior"
    )
    gate.copy_user_info(bottom_central_shell, bottom_central_interior)
    bottom_central_interior.size[1] -= thickness
    bottom_central_interior.size[2] -= thickness

    # union
    interior = gate.solid_union(top_interior, bottom_left_interior, [7 * cm, 0, 0])
    interior = gate.solid_union(
        interior, bottom_central_interior, [0, c + thickness / 2, 0]
    )
    interior = gate.solid_union(interior, bottom_right_interior, [-7 * cm, 0, 0])
    interior = simulation.add_volume_from_solid(interior, f"{name}_interior")
    interior.mother = name
    interior.material = water
    interior.color = blue
    interior.color = [1, 0, 0, 1]

    # central tube in iec_plastic
    cc = simulation.add_volume("Tubs", f"{name}_center_cylinder")
    cc.mother = f"{name}_interior"
    cc.rmax = 2.5 * cm
    cc.rmin = 2.1 * cm
    cc.dz = top_interior.dz
    cc.sphi = 0 * deg
    cc.dphi = 360 * deg
    cc.material = iec_plastic
    cc.translation = [0, 3.5 * cm, 0]
    cc.color = red

    # central tube lung material
    hscc = simulation.add_volume("Tubs", f"{name}_center_cylinder_hole")
    hscc.mother = f"{name}_interior"
    hscc.rmax = 2.1 * cm
    hscc.rmin = 0 * cm
    hscc.dz = top_interior.dz
    hscc.material = iec_lung
    hscc.translation = [0, 3.5 * cm, 0]
    hscc.color = gray

    # spheres
    v = f"{name}_interior"
    iec_add_sphere(
        simulation,
        name,
        v,
        10 * mm,
        1 * mm,
        3 * mm,
        [2.86 * cm, c + 2.39633 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        v,
        13 * mm,
        1 * mm,
        3 * mm,
        [-2.86 * cm, c + 2.39633 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation, name, v, 17 * mm, 1 * mm, 3 * mm, [-5.72 * cm, 3.5 * cm, 3.7 * cm]
    )
    iec_add_sphere(
        simulation,
        name,
        v,
        22 * mm,
        1 * mm,
        3.5 * mm,
        [-2.86 * cm, 8.45367 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        v,
        28 * mm,
        1 * mm,
        3.5 * mm,
        [2.86 * cm, 8.45367 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation, name, v, 37 * mm, 1 * mm, 3.5 * mm, [5.72 * cm, 3.5 * cm, 3.7 * cm]
    )

    return iec


def add_phantom_old(simulation, name="iec"):
    cm = gate.g4_units("cm")
    mm = gate.g4_units("mm")
    deg = gate.g4_units("deg")
    create_material()

    # colors
    white = [1, 1, 1, 1]
    red = [1, 0, 0, 1]
    blue = [0, 0, 1, 1]
    lightblue = [0.4, 0.4, 1, 1]
    gray = [0.5, 0.5, 0.5, 1]
    green = [0, 1, 0, 1]

    # material
    simulation.g4_check_overlap_flag = False

    # main volume
    iec = simulation.add_volume("Tubs", "iec")
    iec.rmax = 17 * cm
    iec.rmin = 0 * cm
    iec.dz = 22 * cm / 2
    iec.sphi = 0 * deg
    iec.dphi = 360 * deg
    iec.material = "G4_AIR"
    iec.color = white

    # ---------------------
    # Upper Half of Phantom

    # Upper outer shell
    uos = simulation.add_volume("Tubs", "upperadius_outer_shell")
    uos.mother = "iec"
    uos.rmax = 15 * cm
    uos.rmin = 14.7 * cm
    uos.dz = 21.4 * cm / 2
    uos.sphi = 0 * deg
    uos.dphi = 180 * deg
    uos.material = iec_plastic
    uos.translation = [0, -3.5 * cm, 0]

    # Upper interior
    ui = simulation.add_volume("Tubs", "upper_interior")
    gate.copy_user_info(uos, ui)
    ui.rmax = uos.rmin
    ui.rmin = 0 * cm
    ui.material = "G4_WATER"

    # iec_plastic Shell Surrounding Lung Insert (Center Cylinder)
    cc = simulation.add_volume("Tubs", "center_cylinder")
    cc.mother = "upper_interior"
    cc.rmax = 2.5 * cm
    cc.rmin = 2.1 * cm
    cc.dz = uos.dz
    cc.sphi = 0 * deg
    cc.dphi = 360 * deg
    cc.material = iec_plastic
    cc.translation = [0, 3.5 * cm, 0]

    # Hollow Space in Central Cylinder
    hscc = simulation.add_volume("Tubs", "center_cylinder_hole")
    hscc.mother = "upper_interior"
    hscc.rmax = 2.1 * cm
    hscc.rmin = 0 * cm
    hscc.dz = uos.dz
    hscc.material = "G4_LUNG_ICRP"
    hscc.translation = [0, 3.5 * cm, 0]

    # Exterior Shell of Upper Half of Phantom

    # Top Side
    ts = simulation.add_volume("Tubs", "top_shell")
    ts.mother = "iec"
    ts.rmax = 15 * cm
    ts.rmin = 0 * cm
    ts.dz = 0.3 * cm / 2
    ts.sphi = 0 * deg
    ts.dphi = 180 * deg
    ts.translation = [0, -3.5 * cm, 10.85 * cm]
    ts.material = iec_plastic

    # bottom side
    bs = simulation.add_volume("Tubs", "bottom_shell")
    gate.copy_user_info(ts, bs)
    bs.translation[2] *= -1

    # Lower left half of phantom
    blos = simulation.add_volume("Tubs", "bottom_left_outer_shell")
    blos.mother = "iec"
    blos.rmax = 8 * cm
    blos.rmin = 7.7 * cm
    blos.dz = 21.4 * cm / 2
    blos.sphi = 270 * deg
    blos.dphi = 90 * deg
    blos.translation = [7 * cm, -3.5 * cm, 0]
    blos.material = iec_plastic

    # Lower Left interior
    lli = simulation.add_volume("Tubs", "lower_left_interior")
    gate.copy_user_info(blos, lli)
    lli.rmax = blos.rmin
    lli.rmin = 0
    lli.material = "G4_WATER"

    # Lower right half of phantom
    bros = simulation.add_volume("Tubs", "bottom_right_outer_shell")
    gate.copy_user_info(blos, bros)
    bros.sphi = 180 * deg
    bros.translation[0] *= -1

    # Lower right interior
    lri = simulation.add_volume("Tubs", "lower_right_interior")
    gate.copy_user_info(lli, lri)
    lri.sphi = 180 * deg
    lri.translation[0] *= -1

    # Bottom box
    bb = simulation.add_volume("Box", "bottom_box")
    bb.size = [14 * cm, 0.3 * cm, 21.4 * cm]
    bb.translation = [0, -11.35 * cm, 0]
    bb.mother = "iec"
    bb.material = iec_plastic

    # Interior box
    ib = simulation.add_volume("Box", "interior_box")
    gate.copy_user_info(bb, ib)
    ib.material = "G4_WATER"
    ib.size[1] = 7.7 * cm
    ib.translation[1] = -7.35 * cm

    # top shell
    ts2 = simulation.add_volume("Tubs", "top_shell2")
    ts2.mother = "iec"
    ts2.rmax = 8 * cm
    ts2.rmin = 0 * cm
    ts2.dz = 0.3 * cm / 2
    ts2.sphi = 270 * deg
    ts2.dphi = 90 * deg
    ts2.material = iec_plastic
    ts2.translation = [7 * cm, -3.5 * cm, 10.85 * cm]

    # top shell
    ts3 = simulation.add_volume("Tubs", "top_shell3")
    gate.copy_user_info(ts2, ts3)
    ts3.sphi = 180 * deg
    ts3.translation[0] *= -1

    # top shell
    ts4 = simulation.add_volume("Box", "top_shell4")
    ts4.mother = "iec"
    ts4.size = [14 * cm, 8 * cm, 0.3 * cm]
    ts4.material = iec_plastic
    ts4.translation = [0 * cm, -7.5 * cm, 10.85 * cm]

    # bottom shell
    bs2 = simulation.add_volume("Tubs", "bottom_shell2")
    gate.copy_user_info(ts2, bs2)
    bs2.translation[2] *= -1

    # bottom shell
    bs3 = simulation.add_volume("Tubs", "bottom_shell3")
    gate.copy_user_info(ts3, bs3)
    bs3.translation[2] *= -1

    # bottom shell
    bs4 = simulation.add_volume("Box", "bottom_shell4")
    gate.copy_user_info(ts4, bs4)
    bs4.translation[2] *= -1

    # spheres
    iec_add_sphere(
        simulation,
        name,
        "interior_box",
        10 * mm,
        1 * mm,
        3 * mm,
        [2.86 * cm, 2.39633 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        "interior_box",
        13 * mm,
        1 * mm,
        3 * mm,
        [-2.86 * cm, 2.39633 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        "upper_interior",
        17 * mm,
        1 * mm,
        3 * mm,
        [-5.72 * cm, 3.5 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        "upper_interior",
        22 * mm,
        1 * mm,
        3.5 * mm,
        [-2.86 * cm, 8.45367 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        "upper_interior",
        28 * mm,
        1 * mm,
        3.5 * mm,
        [2.86 * cm, 8.45367 * cm, 3.7 * cm],
    )
    iec_add_sphere(
        simulation,
        name,
        "upper_interior",
        37 * mm,
        1 * mm,
        3.5 * mm,
        [5.72 * cm, 3.5 * cm, 3.7 * cm],
    )

    # colors
    tree = simulation.volume_manager.build_tree()
    vol = tree[iec.name]
    for v in LevelOrderIter(vol):
        vv = simulation.volume_manager.user_info_volumes[v.name]
        if vv.material == water:
            vv.color = blue
        if vv.material == iec_plastic:
            vv.color = lightblue
        if vv.material == iec_lung:
            vv.color = gray
        if "sphere" in vv.name:
            vv.color = green
        if "capillary" in vv.name:
            vv.color = green

    ts2.color = [1, 0, 0, 1]

    return iec


def iec_add_sphere(sim, name, vol, diam, sph_thick, cap_thick, position):
    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")
    d = f"{(diam / mm):.0f}mm"
    rad = diam / 2

    # interior sphere
    sph = sim.add_volume("Sphere", f"{name}_sphere_{d}")
    sph.mother = vol
    sph.translation = np.array(position)  # need to copy the array!
    sph.rmax = rad
    sph.rmin = 0
    sph.material = "G4_WATER"

    # outer sphere shell
    sphs = sim.add_volume("Sphere", f"{name}_sphere_shell_{d}")
    sphs.mother = vol
    sphs.translation = np.array(position)
    sphs.rmax = rad + sph_thick
    sphs.rmin = rad
    sphs.material = iec_plastic

    # capillary
    cap = sim.add_volume("Tubs", f"{name}_capillary_{d}")
    cap.mother = vol
    cap.translation = np.array(position)
    cap.material = "G4_WATER"
    cap.rmax = 0.25 * cm
    cap.rmin = 0 * cm
    # 21.4/2 = 10.7 interior height (top_interior)
    thickness = 0.3 * cm
    h = 21.4 / 2 * cm - thickness
    cap.dz = (h - 3.7 * cm - rad - sph_thick) / 2.0
    cap.translation[2] = 3.7 * cm + rad + sph_thick + cap.dz

    # capillary outer shell
    caps = sim.add_volume("Tubs", f"{name}_capillary_shell_{d}")
    gate.copy_user_info(cap, caps)
    caps.material = iec_plastic
    caps.rmax = cap_thick
    caps.rmin = cap.rmax


def add_spheres_sources(
    simulation, iec_name, src_name, spheres, activity_Bq_mL, verbose=False
):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    sources = []
    if spheres == "all":
        spheres = spheres_diam
    for sphere, ac in zip(spheres, activity_Bq_mL):
        if sphere in spheres_diam:
            if ac > 0:
                s = add_one_sphere_source(
                    simulation, iec_name, src_name, float(sphere), float(ac)
                )
                sources.append(s)
        else:
            gate.fatal(
                f"Error the sphere of diameter {sphere} does not exists in {spheres_diam}"
            )
    # verbose ?
    if verbose:
        s = dump_spheres_activity(simulation, iec_name, src_name)
        print(s)
    return sources


def dump_spheres_activity(simulation, iec_name, src_name):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    out = ""
    mm = gate.g4_units("mm")
    cm3 = gate.g4_units("cm3")
    Bq = gate.g4_units("Bq")
    BqmL = Bq / cm3
    for diam in spheres_diam:
        d = f"{(diam / mm):.0f}mm"
        sname = f"{src_name}_{iec_name}_{d}"
        if sname not in simulation.source_manager.user_info_sources:
            continue
        src = simulation.get_source_user_info(sname)
        vname = src.mother
        v = simulation.get_volume_user_info(vname)
        s = simulation.get_solid_info(v)
        ac = src.activity
        out += (
            f"{vname:<20} {sname:<20} "
            f"{s.cubic_volume / cm3:10.2f} mL   {ac / Bq:10.2f} Bq   {ac / s.cubic_volume / BqmL:10.2f} Bq/mL\n"
        )
    return out[:-1]


def dump_bg_activity(simulation, iec_name, src_name):
    cm3 = gate.g4_units("cm3")
    Bq = gate.g4_units("Bq")
    BqmL = Bq / cm3
    sname = f"{iec_name}_{src_name}_bg"
    if sname not in simulation.source_manager.user_info_sources:
        return
    src = simulation.get_source_user_info(sname)
    vname = src.mother
    v = simulation.get_volume_user_info(vname)
    s = simulation.get_solid_info(v)
    ac = src.activity
    out = (
        f"{vname:<20} {sname:<20} "
        f"{s.cubic_volume / cm3:10.2f} mL   {ac / Bq:10.2f} Bq   {ac / s.cubic_volume / BqmL:10.2f} Bq/mL"
    )
    return out


def add_one_sphere_source(simulation, iec_name, src_name, diameter, activity_Bq_mL):
    mm = gate.g4_units("mm")
    mL = gate.g4_units("mL")
    d = f"{(diameter / mm):.0f}mm"
    sname = f"{iec_name}_sphere_{d}"

    # compute volume in mL (and check)
    volume_ref = 4 / 3 * np.pi * np.power(diameter / mm / 2, 3) * 0.001
    v = simulation.get_volume_user_info(sname)
    s = simulation.get_solid_info(v)
    volume = s.cubic_volume / mL
    if not math.isclose(volume_ref, volume, rel_tol=1e-7):
        gate.fatal(
            f"Error while estimating the sphere volume {sname}: {volume_ref} vs {volume}"
        )

    source = simulation.add_source("Generic", f"{src_name}_{iec_name}_{d}")
    source.particle = "e+"
    source.energy.type = "F18"
    source.direction.type = "iso"
    source.activity = activity_Bq_mL * s.cubic_volume
    source.position.type = "sphere"
    source.position.radius = diameter / 2 * mm
    source.position.translation = [0, 0, 0]
    source.mother = sname
    return source


def add_background_source(
    simulation, iec_name, src_name, activity_Bq_mL, verbose=False
):
    cm3 = gate.g4_units("cm3")
    # this source is confined only on mother volume, it does not include daughter volumes
    bg = simulation.add_source("Generic", f"{iec_name}_{src_name}_bg")
    bg.mother = f"{iec_name}_interior"
    v = simulation.get_volume_user_info(bg.mother)
    s = simulation.get_solid_info(v)
    # (1 cm3 = 1 mL)
    bg_volume = s.cubic_volume / cm3
    bg.position.type = "box"
    bg.position.size = gate.get_volume_bounding_box_size(simulation, bg.mother)
    bg.position.confine = bg.mother
    bg.particle = "e+"
    bg.energy.type = "F18"
    bg.activity = activity_Bq_mL * s.cubic_volume
    # verbose ?
    if verbose:
        s = dump_bg_activity(simulation, iec_name, src_name)
        print(s)
    return bg


def generate_pos_dir_one_sphere(center, radius, n, rs=np.random):
    """
    This function should be useful to generate conditional data for condGAN.
    It samples the position in a sphere and isotropic direction.
    The center/radius is the center and radius of the sphere
    A numpy array of (n,6) is returned.
    3 first = position
    3 last = direction
    """
    # uniform random vector of size n
    p = generate_pos_one_sphere(center, radius, n, rs)
    # direction
    v = gate.generate_isotropic_directions(n, rs=rs)
    # concat all
    return np.column_stack((p, v))


def generate_pos_one_sphere(center, radius, n, rs=np.random):
    # uniform random vector of size n
    u = rs.uniform(0, 1, size=n)
    r = np.cbrt((u * radius**3))
    phi = rs.uniform(0, 2 * np.pi, n)
    theta = np.arccos(rs.uniform(-1, 1, n))
    # position in cartesian
    x = r * np.sin(theta) * np.cos(phi) + center[0]
    y = r * np.sin(theta) * np.sin(phi) + center[1]
    z = r * np.cos(theta) + center[2]
    # concat all
    return np.column_stack((x, y, z))


def generate_pos_dir_spheres(centers, radius, n_samples, shuffle, rs=np.random):
    """
    This function generate conditional data for condGAN.
    It samples the position in several spheres, with isotropic direction.
    The center/radius are the center and radius of the spheres.
    n_samples is the number of samples per sphere, with a total of n.
    Samples can be shuffled (by default).
    A numpy array of (n,6) is returned.
    """
    cond = None
    for rad, center, n in zip(radius, centers, n_samples):
        # approximate -> if the last one we complete to reach n
        x = generate_pos_dir_one_sphere(center, rad, n, rs=rs)
        if cond is None:
            cond = x
        else:
            cond = np.vstack((cond, x))

    # shuffle
    if shuffle:
        # it seems that permutation is **much** faster than shuffle
        # (checked 2022/06/047 on osx)

        # https://github.com/numpy/numpy/issues/11013
        # sstart = time.time()
        # np.random.shuffle(cond)
        # send = time.time()
        # print(f'shuffle 1 {send - sstart:0.4f} sec')

        # sstart = time.time()
        cond = cond.take(rs.permutation(cond.shape[0]), axis=0)
        # send = time.time()
        # print(f'shuffle 2 {send - sstart:0.4f} sec')

    return cond


def generate_pos_spheres(centers, radius, n_samples, shuffle, rs=np.random):
    """
    Like generate_pos_dir_spheres, but position only
    """
    cond = None
    for rad, center, n in zip(radius, centers, n_samples):
        # approximate -> if the last one we complete to reach n
        x = generate_pos_one_sphere(center, rad, n, rs)
        if cond is None:
            cond = x
        else:
            cond = np.vstack((cond, x))

    # shuffle
    if shuffle:
        cond = cond.take(rs.permutation(cond.shape[0]), axis=0)

    return cond


def get_n_samples_from_ratio(n, ratio):
    """
    For a given proportion of activities (in ratio) and total number of particle n,
    compute the list of particle for each index.
    """
    i = 0
    total = 0
    n_samples = []
    for r in ratio:
        if i == len(ratio) - 1:
            # last one ?
            m = n - total
        else:
            m = int(round(n * r))
        n_samples.append(m)
        total += m
        i += 1
    return n_samples


def compute_sphere_centers_and_volumes(sim, name):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    centers = []
    volumes = []
    mm = gate.g4_units("mm")
    for diam in spheres_diam:
        # retrieve the name of the sphere volume
        d = f"{(diam / mm):.0f}mm"
        vname = f"{name}_sphere_{d}"
        v = sim.get_volume_user_info(vname)
        s = sim.get_solid_info(v)
        # from the solid get the center position
        center = v.translation
        centers.append(center)
        # and the volume
        volumes.append(s.cubic_volume)
    return centers, volumes


def get_default_sphere_centers_and_volumes():
    """
    Global spheres centers in the phantom, to avoid using the phantom in same cases.
    Were computed with 10/06/2022 version.
    No translation. To be recomputed with compute_sphere_centers_and_volumes
    """
    centers = [
        [28.6, -16.0367, 37.0],
        [-28.6, -16.0367, 37.0],
        [-57.2, 35.0, 37.0],
        [-28.6, 84.5367, 37.0],
        [28.6, 84.5367, 37.0],
        [57.2, 35.0, 37.0],
    ]
    volumes = [
        523.5987755982989,
        1150.3465099894627,
        2572.4407845144424,
        5575.279762570685,
        11494.040321933857,
        26521.84878038063,
    ]
    return centers, volumes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# global log level
# create the simulation
sim = gate.Simulation()
print(f"Volumes types: {sim.dump_volume_types()}")

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

# add a material database
sim.add_material_database(pathFile / ".." / "data" / "GateMaterials.db")

#  change world size
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# create a union of several volumes

# first create the solids
b = sim.new_solid("Box", "box")
b.size = [10 * cm, 10 * cm, 10 * cm]
s = sim.new_solid("Sphere", "sphere")
s.rmax = 5 * cm
t = sim.new_solid("Tubs", "t")
t.rmin = 0
t.rmax = 2 * cm
t.dz = 15 * cm

# bool operations
a = gate.solid_union(b, s, [0, 1 * cm, 5 * cm])
a = gate.solid_subtraction(a, t, [0, 1 * cm, 5 * cm])
a = gate.solid_union(a, b, [0, -1 * cm, -5 * cm])  # strange but ok
b = gate.solid_intersection(t, s, [3 * cm, 0, 0])
a = gate.solid_union(a, b, [0, -7 * cm, -5 * cm])

# then add them to a Union, with translation/rotation
rot = Rotation.from_euler("x", 33, degrees=True).as_matrix()
u = sim.add_volume_from_solid(a, "my_stuff")
u.translation = [5 * cm, 5 * cm, 5 * cm]
u.rotation = rot
u.mother = "world"
u.material = "G4_WATER"
u.color = [0, 1, 0, 1]

# create a volume from a solid (not really useful)
u = sim.add_volume_from_solid(s, "test_sph")
u.translation = [-5 * cm, -5 * cm, 1 - 5 * cm]
u.mother = "world"
u.material = "G4_WATER"
u.color = [0, 1, 1, 1]

# default source for tests
source = sim.add_source("Generic", "Default")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.particle = "proton"
source.energy.mono = 240 * MeV
source.position.radius = 1 * cm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 5 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# create G4 objects
print(sim)
sim.initialize()

# explicit check overlap (already performed during initialize)
sim.check_volumes_overlap(verbose=True)

# start simulation
sim.start()

# print results at the end
stats = sim.get_actor("Stats")
print(stats)

# check
# FIXME


gate.test_ok(False)



import opengate as gate
import opengate_core as g4
import re
from box import Box


class MaterialBuilder:
    """
    Manage information (read from a file) to build a Geant4 material.
    Can be a G4Istope, G4Element or G4Material (a C-compound)
    """

    def __init__(self, material_database):
        self.type = "element"
        self.name = None
        self.symbol = None
        self.Zeff = None
        self.Aeff = None
        self.density = None
        self.n = None
        self.state = None
        self.elements = {}
        self.material_database = material_database

    def __del__(self):
        pass

    def __repr__(self):
        u = gate.g4_units("g/mole")
        if self.type == "element":
            s = f"({self.type}) {self.name} ({self.symbol}) Z={self.Zeff} A={self.Aeff / u} g/mole"
        else:
            s = f"({self.type}) {self.name} {self.density} {self.n} {self.elements}"
        return s

    def read_tag(self, s, tag):
        w = s.split("=")
        if w[0].strip() != tag:
            return None
        value = w[1].strip()
        return value

    def read_tag_with_unit(self, s, tag):
        w = s.split("=")
        if w[0].strip() != tag:
            return None
        w = w[1].split()
        value = float(w[0])
        u = gate.g4_units(w[1].strip())
        return value * u

    def read_element(self, f, line):
        self.type = "element"
        s = line.split(":")
        # name
        self.name = s[0]
        s = s[1].strip()
        s = s.split(";")
        # symbol
        self.symbol = self.read_tag(s[0], "S")
        # Z
        self.Zeff = float(self.read_tag(s[1], "Z"))
        # A with units
        self.Aeff = self.read_tag_with_unit(s[2], "A")

    def read_material(self, f, line):
        self.type = "material"
        s = line.split(":")
        # name
        name = s[0]
        if name == "+el":
            gate.fatal(
                f"Error line {line}, missing elements for the previous material ?"
            )
        self.name = name
        s = s[1].split(";")
        if len(s) != 3 and len(s) != 2:
            gate.fatal(f"Error while parsing material {self.name}, line {line}")
        # density
        self.density = self.read_tag_with_unit(s[0], "d")
        if not self.density:
            gate.fatal(
                f"Error while parsing material {self.name}, line {line}\n"
                f'Expected density with "d=XXX"'
            )
        # nb of elements
        self.n = int(self.read_tag(s[1], "n"))
        gcm3 = gate.g4_units("g/cm3")
        # state
        if len(s) > 2:
            self.state = self.read_tag(s[2], "state")
            if self.state:
                self.state = self.state.lower()
        # elements
        elems = []
        for e in range(self.n):
            ee = self.read_one_element(f)
            elems.append(ee)
        # update the fraction
        n_is_used = elems[0].n
        for ee in elems:
            if ee.n and not n_is_used:
                gate.fatal(
                    f'Error, some elements used "n" while other used "f", {self}'
                )
        if not n_is_used:
            total = 0
            for ee in elems:
                total += ee.f
            for ee in elems:
                ee.f = ee.f / total

    def read_one_element(self, f):
        line = f.readline().strip()
        if line[:4] != "+el:":
            gate.fatal(
                f'Error, expect "+el:" at the beginning of this line: {line}\n'
                f" while parsing the material {self.name}"
            )
            return
        s = line.split("+el:")
        s = re.split(";|,", s[1])
        if len(s) != 2:
            gate.fatal(
                f"Error while reading the line: {line} \n"
                f'Expected "name=" ; "n=" or "f="'
            )
        elname = self.read_tag(s[0], "name")
        if elname == "auto":
            elname = self.name
        if not elname:
            gate.fatal(
                f"Error reading line {line} \n during the elements of material {self.name}"
            )
        f = None
        n = self.read_tag(s[1], "n")
        if not n:
            f = float(self.read_tag(s[1], "f"))
        else:
            n = int(n)
        e = Box({"name": elname, "n": n, "f": f})
        self.elements[elname] = e
        return e

    def build(self):
        if self.type == "element":
            return self.build_element()
        if self.type == "isotope":
            return self.build_isotope()
        if self.type == "material":
            return self.build_material()
        gate.fatal(f"Error, material type unknown: {self.type}")

    def build_isotope(self):
        print("build_isotope")
        gate.fatal(f"Not yet implemented")

    def build_element(self):
        m = g4.G4Element(self.name, self.symbol, self.Zeff, self.Aeff)
        # FIXME alternative with Build an element from isotopes via AddIsotope ?
        return m

    def build_material(self):
        n = len(self.elements)
        switcher = {
            None: g4.G4State.kStateUndefined,
            "solid": g4.G4State.kStateSolid,
            "liquid": g4.G4State.kStateLiquid,
            "gaz": g4.G4State.kStateGas,
        }
        state = switcher.get(self.state, f"Invalid material state {self.state}")
        # default temp
        kelvin = gate.g4_units("kelvin")
        temp = 293.15 * kelvin
        # default pressure
        atmosphere = gate.g4_units("atmosphere")
        pressure = 1 * atmosphere
        # create material
        m = g4.G4Material(self.name, self.density, n, state, temp, pressure)
        # warning: cannot print while all elements are not added
        for elem in self.elements:
            b = self.material_database.FindOrBuildElement(elem)
            if self.elements[elem].f is None:
                m.AddElement_n(b, self.elements[elem].n)
            else:
                m.AddElement_f(b, self.elements[elem].f)
        return m



import opengate as gate
import opengate_core as g4


class BoxVolume(gate.VolumeBase):
    type_name = "Box"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        cm = gate.g4_units("cm")
        user_info.size = [10 * cm, 10 * cm, 10 * cm]

    def build_solid(self):
        u = self.user_info
        return g4.G4Box(u.name, u.size[0] / 2.0, u.size[1] / 2.0, u.size[2] / 2.0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# global log level
# create the simulation
sim = gate.Simulation()
print(f"Volumes types: {sim.dump_volume_types()}")

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

# add a material database
sim.add_material_database(pathFile / ".." / "data" / "GateMaterials.db")

#  change world size
m = gate.g4_units("m")
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]

# add a simple volume
waterbox = sim.add_volume("Box", "Waterbox")
cm = gate.g4_units("cm")
waterbox.size = [60 * cm, 60 * cm, 60 * cm]
waterbox.translation = [0 * cm, 0 * cm, 35 * cm]
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]  # blue

# another (child) volume with rotation
mm = gate.g4_units("mm")
sheet = sim.add_volume("Box", "Sheet")
sheet.size = [30 * cm, 30 * cm, 2 * mm]
sheet.mother = "Waterbox"
sheet.material = "Lead"
r = Rotation.from_euler("x", 33, degrees=True)
center = [0 * cm, 0 * cm, 10 * cm]
t = gate.get_translation_from_rotation_with_center(r, center)
sheet.rotation = r.as_matrix()
sheet.translation = t + [0 * cm, 0 * cm, -18 * cm]
sheet.color = [1, 0, 0, 1]  # red

# A sphere
sph = sim.add_volume("Sphere", "mysphere")
sph.rmax = 5 * cm
sph.mother = "Waterbox"
sph.translation = [0 * cm, 0 * cm, -8 * cm]
sph.material = "Lung"
sph.color = [0.5, 1, 0.5, 1]  # kind of green
sph.toto = "nothing"  # ignored, should raise a warning

# A ...thing ?
trap = sim.add_volume("Trap", "mytrap")
trap.mother = "Waterbox"
trap.translation = [0, 0, 15 * cm]
trap.material = "G4_LUCITE"

# default source for tests
source = sim.add_source("Generic", "Default")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.particle = "proton"
source.energy.mono = 240 * MeV
source.position.radius = 1 * cm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 2500 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# run timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [
    [0, 0.5 * sec]
    # ,[0.5 * sec, 1.2 * sec]
]

# create G4 objects
print(sim)
sim.initialize()

# explicit check overlap (already performed during initialize)
sim.check_volumes_overlap(verbose=True)

# print info material db
dbn = sim.dump_material_database_names()
mnist = sim.dump_material_database("NIST")
mdb = sim.dump_material_database(pathFile / ".." / "data" / "GateMaterials.db")
dm = sim.dump_defined_material()
print("Material info:")
print("\t databases    :", dbn)
print("\t mat in NIST  :", len(mnist), mnist)
print("\t mat in db    :", mdb)
print("\t defined mat  :", dm)

assert dbn == [pathFile / ".." / "data" / "GateMaterials.db", "NIST"]
assert len(mnist) == 308
assert mdb == [
    "Vacuum",
    "Aluminium",
    "Uranium",
    "Silicon",
    "Germanium",
    "Yttrium",
    "Gadolinium",
    "Lutetium",
    "Tungsten",
    "Lead",
    "Bismuth",
    "NaI",
    "NaITl",
    "PWO",
    "BGO",
    "LSO",
    "Plexiglass",
    "GSO",
    "LuAP",
    "YAP",
    "Water",
    "Quartz",
    "Breast",
    "Air",
    "Glass",
    "Scinti-C9H10",
    "LuYAP-70",
    "LuYAP-80",
    "Plastic",
    "CZT",
    "Lung",
    "Polyethylene",
    "PVC",
    "SS304",
    "PTFE",
    "LYSO",
    "Body",
    "Muscle",
    "LungMoby",
    "SpineBone",
    "RibBone",
    "Adipose",
    "Blood",
    "Heart",
    "Kidney",
    "Liver",
    "Lymph",
    "Pancreas",
    "Intestine",
    "Skull",
    "Cartilage",
    "Brain",
    "Spleen",
    "Testis",
    "PMMA",
]
assert dm == ["G4_AIR", "G4_WATER", "Lead", "Lung", "G4_LUCITE"]

# verbose
sim.apply_g4_command("/tracking/verbose 0")
# sim.g4_com("/run/verbose 2")
# sim.g4_com("/event/verbose 2")
# sim.g4_com("/tracking/verbose 1")

# start simulation
sim.start()

# print results at the end
stats = sim.get_actor("Stats")
print(stats)

# check
assert len(sim.dump_defined_material()) == 5
stats_ref = gate.SimulationStatisticsActor()
c = stats_ref.counts
c.run_count = 1
c.event_count = 1280
c.track_count = 17034  # 25668
c.step_count = 78096  # 99465
# stats_ref.pps = 506.6
sec = gate.g4_units("second")
c.duration = 2.5267 * sec
print("-" * 80)
is_ok = gate.assert_stats(stats, stats_ref, 0.15)

gate.test_ok(is_ok)



from box import Box
import opengate as gate
import opengate_core as g4
from anytree import Node

""" Global name for the world volume"""
__world_name__ = "world"


class VolumeManager(g4.G4VUserDetectorConstruction):
    """
    Implementation of G4VUserDetectorConstruction.
    In 'Construct' function, build all volumes in the scene.
    Keep a list of solid, logical volumes, physical volumes, materials.
    """

    def __init__(self, simulation):
        """
        Class that store geometry description.
        self.geometry is the dict that describes all parameters
        self.geometry_tree is the volumes sorted as a tree
        other are g4 objects
        """
        g4.G4VUserDetectorConstruction.__init__(self)
        self.simulation = simulation
        # list of all user_info describing the volumes
        self.user_info_volumes = {}  # user info only
        self.volumes_tree = None
        # list of all build volumes (only after initialization
        self.volumes = {}
        self.is_constructed = False
        # G4 elements are stored to avoid auto destruction
        # and allows access
        self.g4_materials = Box()
        # Materials databases
        self.g4_NistManager = None
        self.material_databases = {}
        self.element_names = []
        self.material_names = []

    def __del__(self):
        pass

    def __str__(self):
        s = (
            f"{len(self.user_info_volumes)} volumes,"
            f" {len(self.volumes)} are constructed"
        )
        return s

    def get_volume_info(self, name):
        if name not in self.user_info_volumes:
            gate.fatal(
                f"The volume {name} is not in the current "
                f"list of volumes: {self.user_info_volumes}"
            )
        return self.user_info_volumes[name]

    def get_volume(self, name, check_initialization=True):
        if check_initialization and not self.is_constructed:
            gate.fatal(f"Cannot get_volume before initialization")
        if name not in self.volumes:
            gate.fatal(
                f"The volume {name} is not in the current "
                f"list of volumes: {self.volumes}"
            )
        return self.volumes[name]

    def new_solid(self, solid_type, name):
        if solid_type == "Boolean":
            gate.fatal(f"Cannot create solid {solid_type}")
        # Create a UserInfo for a volume
        u = gate.UserInfo("Volume", solid_type, name)
        # remove unused keys: object, etc (it's a solid, not a volume)
        VolumeManager._pop_keys_unused_by_solid(u)
        return u

    def get_solid_info(self, user_info):
        """
        Temporary build a solid from the user info, in order to retrieve information (volume etc).
        Can be used *before* initialization
        """
        vol = gate.new_element(user_info, self.simulation)
        vol = vol.build_solid()
        r = Box()
        r.cubic_volume = vol.GetCubicVolume()
        r.surface_area = vol.GetSurfaceArea()
        pMin = g4.G4ThreeVector()
        pMax = g4.G4ThreeVector()
        vol.BoundingLimits(pMin, pMax)
        r.bounding_limits = [pMin, pMax]
        return r

    def _pop_keys_unused_by_solid(user_info):
        # remove unused keys: object, etc (it's a solid, not a volume)
        u = user_info.__dict__
        u.pop("mother", None)
        u.pop("translation", None)
        u.pop("color", None)
        u.pop("rotation", None)
        u.pop("material", None)

    def add_volume(self, vol_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.user_info_volumes, name)
        # initialize the user_info
        v = gate.UserInfo("Volume", vol_type, name)
        # add to the list
        self.user_info_volumes[name] = v
        # FIXME  NOT CLEAR --> here ? or later
        # create a region for the physics cuts
        # user will be able to set stuff like :
        # pm.production_cuts.my_volume.gamma = 1 * mm
        pm = self.simulation.get_physics_user_info()
        cuts = pm.production_cuts
        cuts[name] = Box()
        # return the info
        return v

    def add_volume_from_solid(self, solid, name):
        v = None
        for op in gate.bool_operators:
            try:
                if op in solid:
                    v = self.add_volume("Boolean", name)
                    v.solid = solid
            except:
                pass
        if not v:
            v = self.add_volume(solid.type_name, name)
            # copy the parameters of the solid
            gate.copy_user_info(solid, v)
        return v

    def add_material_database(self, filename, name):
        if not name:
            name = filename
        if name in self.material_databases:
            gate.fatal(f'Database "{name}" already exist.')
        db = gate.MaterialDatabase(filename, self.material_databases)
        self.material_databases[name] = db

    def Construct(self):
        """
        Override the Construct method from G4VUserDetectorConstruction
        """
        if self.is_constructed:
            gate.fatal("Cannot construct volumes, it has been already done.")

        # tree re-order
        self.check_geometry()
        self.volumes_tree = self.build_tree()

        # default material database: NIST
        self.g4_NistManager = g4.G4NistManager.Instance()
        self.material_databases["NIST"] = self.g4_NistManager
        self.element_names = self.g4_NistManager.GetNistElementNames()
        self.material_names = self.g4_NistManager.GetNistMaterialNames()

        # check for duplicate material names
        # (not sure needed)
        for db in self.material_databases:
            if db == "NIST":
                continue
            for m in self.material_databases[db].material_builders:
                if m in self.material_names:
                    gate.warning(
                        f"Error in db {db}, the material {m} is already defined. Ignored."
                    )
                else:
                    self.material_names.append(m)
            for m in self.material_databases[db].element_builders:
                if m in self.element_names:
                    gate.warning(
                        f"Error in db {db}, the element {m} is already defined. Ignored."
                    )
                else:
                    self.element_names.append(m)

        # build all real volumes object
        for vu in self.user_info_volumes.values():
            # create the volume
            vol = gate.new_element(vu, self.simulation)
            # construct the G4 Volume
            vol.construct(self)
            if len(vol.g4_physical_volumes) == 0:
                vol.g4_physical_volumes.append(vol.g4_physical_volume)
            # keep the volume
            self.volumes[vu.name] = vol

        # return self.g4_physical_volumes.world
        self.is_constructed = True
        return self.volumes[gate.__world_name__].g4_physical_volume

    def dump(self):
        self.check_geometry()
        self.volumes_tree = self.build_tree()
        s = f"Number of volumes: {len(self.user_info_volumes)}"
        s += "\n" + self.dump_tree()
        for vol in self.user_info_volumes.values():
            s += gate.indent(2, f"\n{vol}")
        return s

    def dump_tree(self):
        self.volumes_tree = self.build_tree()
        info = {}
        for v in self.user_info_volumes.values():
            info[v.name] = v
        return gate.pretty_print_tree(self.volumes_tree, info)

    def dump_defined_material(self, level):
        table = g4.G4Material.GetMaterialTable
        if level == 0:
            names = [m.GetName() for m in table]
            return names
        return table

    def check_geometry(self):
        names = {}
        for v in self.volumes:
            vol = self.volumes[v].user_info

            # volume must have a name
            if "_name" not in vol.__dict__:
                gate.fatal(f"Volume is missing a 'name' : {vol}")

            # volume name must be geometry name
            if v != vol.name:
                gate.fatal(
                    f"Volume named '{v}' in geometry has a different name : {vol}"
                )

            if vol.name in names:
                gate.fatal(f"Two volumes have the same name '{vol.name}' --> {self}")
            names[vol.name] = True

            # volume must have a mother, default is gate.__world_name__
            if "mother" not in vol.__dict__:
                vol.mother = gate.__world_name__

            # volume must have a material
            if "material" not in vol.__dict__:
                gate.fatal(f"Volume is missing a 'material' : {vol}")
                # vol.material = 'air'

    def build_tree(self):
        # world is needed as the root
        if gate.__world_name__ not in self.user_info_volumes:
            s = f"No world in geometry = {self.user_info_volumes}"
            gate.fatal(s)

        # build the root tree (needed)
        tree = {gate.__world_name__: Node(gate.__world_name__)}
        already_done = {gate.__world_name__: True}

        # build the tree
        for vol in self.user_info_volumes.values():
            if vol.name in already_done:
                continue
            self._add_volume_to_tree(already_done, tree, vol)

        return tree

    def check_overlaps(self, verbose):
        for v in self.volumes.values():
            for w in v.g4_physical_volumes:
                try:
                    b = w.CheckOverlaps(1000, 0, verbose, 1)
                    if b:
                        gate.fatal(
                            f'Some volumes overlap the volume "{v}". \n'
                            f"Consider using G4 verbose to know which ones. \n"
                            f"Aborting."
                        )
                except:
                    pass
                    # gate.warning(f'do not check physical volume {w}')

    def find_or_build_material(self, material):
        # loop on all databases
        found = False
        mat = None
        for db_name in self.material_databases:
            db = self.material_databases[db_name]
            m = db.FindOrBuildMaterial(material)
            if m and not found:
                found = True
                mat = m
                break
        if not found:
            gate.fatal(f"Cannot find the material {material}")
        # need an object to store the material without destructor
        self.g4_materials[material] = mat
        return mat

    # G4 overloaded
    def ConstructSDandField(self):
        # This function is called in MT mode
        self.simulation.actor_manager.register_sensitive_detectors()

    def _add_volume_to_tree(self, already_done, tree, vol):
        # check if mother volume exists
        if vol.mother not in self.user_info_volumes:
            gate.fatal(
                f"Cannot find a mother volume named '{vol.mother}', for the volume {vol}"
            )

        already_done[vol.name] = "in_progress"
        m = self.user_info_volumes[vol.mother]

        # check for the cycle
        if m.name not in already_done:
            self._add_volume_to_tree(already_done, tree, m)
        else:
            if already_done[m.name] == "in_progress":
                s = f"Error while building the tree, there is a cycle ? "
                s += f"\n volume is {vol}"
                s += f"\n parent is {m}"
                gate.fatal(s)

        # get the mother branch
        p = tree[m.name]

        # check not already exist
        if vol.name in tree:
            s = f"Node already exist in tree {vol.name} -> {tree}"
            s = s + f"\n Probably two volumes with the same name ?"
            gate.fatal(s)

        # create the node
        n = Node(vol.name, parent=p)
        tree[vol.name] = n
        already_done[vol.name] = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

# set log level
# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.verbose_level = gate.DEBUG
ui.running_verbose_level = 0  # gate.EVENT
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1
gate.log.debug(ui)

cm = gate.g4_units("cm")

# add a simple volume
waterbox = sim.add_volume("Box", "Waterbox")
waterbox.size = [20 * cm, 20 * cm, 20 * cm]
waterbox.translation = [0 * cm, 0 * cm, 15 * cm]
waterbox.material = "G4_WATER"

# default source for tests
MeV = gate.g4_units("MeV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("second")
source1 = sim.add_source("Generic", "source1")
source1.particle = "proton"
source1.energy.mono = 150 * MeV
source1.position.radius = 10 * mm
source1.direction.type = "momentum"
source1.direction.momentum = [0, 0, 1]
source1.n = 1000

source2 = sim.add_source("Generic", "source2")
source2.particle = "proton"
source2.energy.mono = 120 * MeV
source2.position.radius = 5 * mm
source2.activity = 1000 * Bq  # 25 + 50 + 100
source2.direction.type = "momentum"
source2.direction.momentum = [0, 0, 1]
source2.start_time = 0.25 * sec

source3 = sim.add_source("Generic", "source3")
source3.particle = "proton"
source3.energy.mono = 150 * MeV
source3.position.radius = 10 * mm
source3.n = 1200
source3.start_time = 0.50 * sec
source3.direction.type = "momentum"
source3.direction.momentum = [0, 0, 1]
source3.toto = 120  # raise a warning

# Expected total of events
# 100 + 175 + 120 = 395

# debug: uncomment to remove one source
# sim.source_manager.sources.pop('source1')
# sim.source_manager.sources.pop('source2')
# sim.source_manager.sources.pop('source3')

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# run timing test #1
sec = gate.g4_units("second")
sim.run_timing_intervals = [
    [0, 0.5 * sec],
    [0.5 * sec, 1.0 * sec],
    # Watch out : there is (on purpose) a 'hole' in the timeline
    [1.5 * sec, 2.5 * sec],
]

# create G4 objects
sim.initialize()
print(sim.dump_sources())

# start simulation
sim.start()

stats = sim.get_actor("Stats")
print(stats)

stats_ref = gate.SimulationStatisticsActor()
c = stats_ref.counts
c.run_count = 3
c.event_count = 3900
c.track_count = 18792  # 56394
c.step_count = 133291  # 217234
# stats_ref.pps = 4059.6 3 3112.2
c.duration = 1 / 4059.6 * 3900 * sec
print("-" * 80)
is_ok = gate.assert_stats(stats, stats_ref, 0.18)

gate.test_ok(is_ok)



from opengate import log
import opengate as gate
import time
import random
import sys
from .ExceptionHandler import *


class Simulation:
    """
    Main class that store and build a simulation.
    """

    def __init__(self, name="simulation"):
        """
        Constructor. Main members are:
        - managers of volumes, sources and actors
        - Geant4 objects that will be build during initialisation (start with g4_)
        - some internal variables
        """
        self.name = name

        # user's defined parameters
        self.user_info = gate.SimulationUserInfo(self)

        # main managers
        self.volume_manager = gate.VolumeManager(self)
        self.source_manager = gate.SourceManager(self)
        self.actor_manager = gate.ActorManager(self)
        self.physics_manager = gate.PhysicsManager(self)
        self.filter_manager = gate.FilterManager(self)
        self.action_manager = None  # will be created later (need source)

        # G4 elements
        self.g4_RunManager = None
        self.g4_PhysList = None
        self.g4_HepRandomEngine = None
        self.g4_ui = None
        self.g4_exception_handler = None

        # internal state
        self.is_initialized = False
        self.run_timing_intervals = None
        self.ui_session = None
        self.actual_random_seed = None

        # default elements
        self._default_parameters()

    def __del__(self):
        # set verbose to zero before destructor to avoid the final message
        if getattr(self, "g4_RunManager", False):
            self.g4_RunManager.SetVerboseLevel(0)
        pass

    def __str__(self):
        i = "initialized"
        if not self.is_initialized:
            i = f"not {i}"
        s = (
            f"Simulation name: {self.name} ({i})\n"
            f"Geometry       : {self.volume_manager}\n"
            f"Physics        : {self.physics_manager}\n"
            f"Sources        : {self.source_manager}\n"
            f"Actors         : {self.actor_manager}"
        )
        return s

    def _default_parameters(self):
        """
        Internal use.
        Build default elements: verbose, World, seed, physics, etc.
        """
        # World volume
        w = self.add_volume("Box", gate.__world_name__)
        w.mother = None
        m = gate.g4_units("meter")
        w.size = [3 * m, 3 * m, 3 * m]
        w.material = "G4_AIR"
        # run timing
        sec = gate.g4_units("second")
        self.run_timing_intervals = [
            [0 * sec, 1 * sec]
        ]  # a list of begin-end time values

    def dump_sources(self):
        return self.source_manager.dump()

    def dump_source_types(self):
        s = f""
        for t in gate.source_builders:
            s += f"{t} "
        return s

    def dump_volumes(self):
        return self.volume_manager.dump()

    def dump_volume_types(self):
        s = f""
        for t in gate.volume_builders:
            s += f"{t} "
        return s

    def dump_actors(self):
        return self.actor_manager.dump()

    def dump_actor_types(self):
        s = f""
        for t in gate.actor_builders:
            s += f"{t} "
        return s

    def dump_material_database_names(self):
        return list(self.volume_manager.material_databases.keys())

    def dump_material_database(self, db, level=0):
        if db not in self.volume_manager.material_databases:
            gate.fatal(
                f'Cannot find the db "{db}" in the '
                f"list: {self.dump_material_database_names()}"
            )
        thedb = self.volume_manager.material_databases[db]
        if db == "NIST":
            return thedb.GetNistMaterialNames()
        return thedb.dump_materials(level)

    def dump_defined_material(self, level=0):
        if not self.is_initialized:
            gate.fatal(f"Cannot dump defined material before initialisation")
        return self.volume_manager.dump_defined_material(level)

    def initialize(self):
        """
        Build the main geant4 objects and initialize them.
        """
        # shorter code
        ui = self.user_info

        # g4 verbose
        self.initialize_g4_verbose()

        # check multithreading
        mt = g4.GateInfo.get_G4MULTITHREADED()
        if ui.number_of_threads > 1 and not mt:
            gate.fatal(
                "Cannot use multi-thread, opengate_core was not compiled with Geant4 MT"
            )

        # check if RunManager already exists (it should not)
        if ui.number_of_threads > 1 or ui.force_multithread_mode:
            rm = g4.G4MTRunManager.GetRunManager()
        else:
            rm = g4.G4RunManager.GetRunManager()
        if rm:
            s = f"Cannot create a Simulation, the G4RunManager already exist."
            gate.fatal(s)

        # init random engine (before the MTRunManager creation)
        self.initialize_random_engine()

        # create the RunManager
        if ui.number_of_threads > 1 or ui.force_multithread_mode:
            log.info(
                f"Simulation: create MTRunManager with {ui.number_of_threads} threads"
            )
            rm = g4.G4MTRunManager()
            rm.SetNumberOfThreads(ui.number_of_threads)
        else:
            log.info("Simulation: create RunManager")
            rm = g4.G4RunManager()
        self.g4_RunManager = rm
        self.g4_RunManager.SetVerboseLevel(ui.g4_verbose_level)

        # Cannot be initialized two times (ftm)
        if self.is_initialized:
            gate.fatal("Simulation already initialized. Abort")

        # create the handler for the exception
        self.g4_exception_handler = ExceptionHandler()

        # check run timing
        gate.assert_run_timing(self.run_timing_intervals)

        # geometry
        log.info("Simulation: initialize Geometry")
        self.g4_RunManager.SetUserInitialization(self.volume_manager)

        # phys
        log.info("Simulation: initialize Physics")
        self.physics_manager.initialize()
        self.g4_RunManager.SetUserInitialization(self.physics_manager.g4_physic_list)

        # sources
        log.info("Simulation: initialize Source")
        self.source_manager.run_timing_intervals = self.run_timing_intervals
        self.source_manager.initialize(self.run_timing_intervals)

        # action
        log.info("Simulation: initialize Actions")
        self.action_manager = gate.ActionManager(self.source_manager)
        self.g4_RunManager.SetUserInitialization(self.action_manager)

        # Actors initialization (before the RunManager Initialize)
        self.actor_manager.create_actors(self.action_manager)
        self.source_manager.initialize_actors(self.actor_manager.actors)

        # Initialization
        log.info("Simulation: initialize G4RunManager")
        self.g4_RunManager.Initialize()
        self.is_initialized = True

        # Physics initialization
        self.physics_manager.initialize_cuts()

        # Actors initialization
        log.info("Simulation: initialize Actors")
        self.actor_manager.initialize()

        # Check overlaps
        if ui.check_volumes_overlap:
            log.info("Simulation: check volumes overlap")
            self.check_volumes_overlap(verbose=False)

        # Register sensitive detector.
        # if G4 was compiled with MT (regardless it is used or not)
        # ConstructSDandField (in VolumeManager) will be automatically called
        if not g4.GateInfo.get_G4MULTITHREADED():
            gate.warning("DEBUG Register sensitive detector in no MT mode")
            self.actor_manager.register_sensitive_detectors()

    def apply_g4_command(self, command):
        """
        For the moment, only use it *after* runManager.Initialize
        """
        if not self.is_initialized:
            gate.fatal(f"Please, use g4_apply_command *after* simulation.initialize()")
        self.g4_ui = g4.G4UImanager.GetUIpointer()
        self.g4_ui.ApplyCommand(command)

    def start(self):
        """
        Start the simulation. The runs are managed in the SourceManager.
        """
        if not self.is_initialized:
            gate.fatal('Use "initialize" before "start"')
        log.info("-" * 80 + "\nSimulation: START")

        # FIXME check run_timing_intervals

        # visualisation should be initialized *after* other initializations
        # FIXME self._initialize_visualisation()

        # actor: start simulation (only the master thread)
        self.actor_manager.start_simulation()

        # go !
        start = time.time()
        self.source_manager.start()
        end = time.time()

        # actor: stop simulation (only the master thread)
        self.actor_manager.stop_simulation()

        # this is the end
        log.info(
            f"Simulation: STOP. Run: {len(self.run_timing_intervals)}. "
            # f'Events: {self.source_manager.total_events_count}. '
            f"Time: {end - start:0.1f} seconds.\n"
            + f"-" * 80
        )

    def initialize_random_engine(self):
        engine_name = self.user_info.random_engine
        self.g4_HepRandomEngine = None
        if engine_name == "MixMaxRng":
            self.g4_HepRandomEngine = g4.MixMaxRng()
        if engine_name == "MersenneTwister":
            self.g4_HepRandomEngine = g4.MTwistEngine()
        if not self.g4_HepRandomEngine:
            s = f"Cannot find the random engine {engine_name}\n"
            s += f"Use: MersenneTwister or MixMaxRng"
            gate.fatal(s)

        # set the random engine
        g4.G4Random.setTheEngine(self.g4_HepRandomEngine)
        if self.user_info.random_seed == "auto":
            self.actual_random_seed = random.randrange(sys.maxsize)
        else:
            self.actual_random_seed = self.user_info.random_seed

        # set the seed
        g4.G4Random.setTheSeed(self.actual_random_seed, 0)

    def initialize_g4_verbose(self):
        # For an unknown reason, when verbose_level == 0, there are some
        # additional print after the G4RunManager destructor. So we default at 1
        ui = None
        if not self.user_info.g4_verbose:
            # no Geant4 output
            ui = gate.UIsessionSilent()
        else:
            # Geant4 output with color
            ui = gate.UIsessionVerbose()
        # it is also possible to set ui=None for 'default' output
        self.set_g4_ui_output(ui)

    def set_g4_ui_output(self, ui_session):
        # we must keep a ref to ui_session
        self.ui_session = ui_session
        # we must keep a ref to ui_manager
        self.g4_ui = g4.G4UImanager.GetUIpointer()
        self.g4_ui.SetCoutDestination(ui_session)

    @property
    def world(self):
        return self.get_volume_user_info(gate.__world_name__)

    def get_volume_user_info(self, name):
        v = self.volume_manager.get_volume_info(name)
        return v

    def get_all_volumes_user_info(self):
        return self.volume_manager.user_info_volumes

    def get_solid_info(self, user_info):
        return self.volume_manager.get_solid_info(user_info)

    def get_source_user_info(self, name):
        return self.source_manager.get_source_info(name)

    def get_source(self, name):
        return self.source_manager.get_source(name)

    def get_source_MT(self, name, thread):
        return self.source_manager.get_source_MT(name, thread)

    def get_actor_user_info(self, name):
        s = self.actor_manager.get_actor_user_info(name)
        return s

    def get_actor(self, name):
        if not self.is_initialized:
            gate.fatal(f"Cannot get an actor before initialization")
        return self.actor_manager.get_actor(name)

    def get_physics_user_info(self):
        return self.physics_manager.user_info

    def set_cut(self, volume_name, particle, value):
        self.physics_manager.set_cut(volume_name, particle, value)

    def set_physics_list(self, pl):
        p = self.get_physics_user_info()
        p.physics_list_name = pl

    def new_solid(self, solid_type, name):
        return self.volume_manager.new_solid(solid_type, name)

    def add_volume(self, solid_type, name):
        return self.volume_manager.add_volume(solid_type, name)

    def add_volume_from_solid(self, solid, name):
        return self.volume_manager.add_volume_from_solid(solid, name)

    def add_source(self, source_type, name):
        return self.source_manager.add_source(source_type, name)

    def add_actor(self, actor_type, name):
        return self.actor_manager.add_actor(actor_type, name)

    def add_filter(self, filter_type, name):
        return self.filter_manager.add_filter(filter_type, name)

    def add_material_database(self, filename, name=None):
        self.volume_manager.add_material_database(filename, name)

    def check_volumes_overlap(self, verbose=True):
        if not self.is_initialized:
            gate.fatal(
                f"Cannot check overlap: the simulation must be initialized before"
            )
        # FIXME: later, allow to bypass this check ?
        # FIXME: How to manage the verbosity ?
        b = self.user_info.g4_verbose
        self.user_info.g4_verbose = True
        self.initialize_g4_verbose()
        self.volume_manager.check_overlaps(verbose)
        self.user_info.g4_verbose = b
        self.initialize_g4_verbose()



import opengate as gate
import opengate_core as g4


class ActionManager(g4.G4VUserActionInitialization):
    """
    Main object to manage all actions during a simulation.
    """

    def __init__(self, source):
        g4.G4VUserActionInitialization.__init__(self)
        # List of G4 source managers (one per thread)
        self.g4_PrimaryGenerator = []
        # The main G4 source manager
        self.g4_main_PrimaryGenerator = None
        # The py source manager
        self.source_manager = source
        # Lists of elements to prevent destruction
        self.g4_RunAction = []
        self.g4_EventAction = []
        self.g4_TrackingAction = []

    def __del__(self):
        pass

    def BuildForMaster(self):
        # This function is call only in MT mode, for the master thread
        if not self.g4_main_PrimaryGenerator:
            self.g4_main_PrimaryGenerator = self.source_manager.build()

    def Build(self):
        # In MT mode the same method is invoked
        # for each worker thread, so all user action classes
        # are defined thread-locally.

        # If MT is not enabled, need to create the main source
        if not self.g4_main_PrimaryGenerator:
            p = self.g4_main_PrimaryGenerator = self.source_manager.build()
        else:
            # else create a source for each thread
            p = self.source_manager.create_g4_source_manager()

        self.SetUserAction(p)
        self.g4_PrimaryGenerator.append(p)

        # set the actions for Run
        ra = g4.GateRunAction(p)
        self.SetUserAction(ra)
        self.g4_RunAction.append(ra)

        # set the actions for Event
        ea = g4.GateEventAction()
        self.SetUserAction(ea)
        self.g4_EventAction.append(ea)

        # set the actions for Track
        ta = g4.GateTrackingAction()
        self.SetUserAction(ta)
        self.g4_TrackingAction.append(ta)



import opengate as gate


def info_timing(i):
    s = f'[ {gate.g4_best_unit(i[0], "Time")}; {gate.g4_best_unit(i[1], "Time")}]'
    return s


def assert_run_timing(run_timing_intervals):
    if len(run_timing_intervals) < 0:
        gate.fatal(
            f"run_timing_interval must be an array "
            f"with at least one element, while it is: {run_timing_intervals}"
        )
    previous_end = run_timing_intervals[0][0]
    for i in run_timing_intervals:
        if len(i) != 2:
            gate.fatal(
                f"This run_timing_interval must have "
                f"exactly one start and one end time, while it is: {i}"
            )
        if i[0] > i[1]:
            gate.fatal(f"Start time must be lower or equal than end time: {i}")
        if i[0] < previous_end:
            gate.fatal(
                f"Start time must be larger or equal than previous end time: {i}, previous: {previous_end}"
            )
        previous_end = i[1]


def info_run_timing(sim: gate.Simulation):
    rti = sim.run_timing_intervals
    s = f"Number of runs: {len(rti)}"
    nr = 0
    for i in rti:
        a = f"\nRun {nr}: {info_timing(i)}"
        s += gate.indent(2, a)
        nr += 1
    return s


def range_timing(start, end, n):
    """
    Return a list of n time intervals, from start to end
    e.g. range_timing(0, 1, 10)
    => [0, 0.1], [0.1, 0.2], [0.2, 0.3], [0.3, 0.4], [0.4, 0.5],
    [0.5, 0.6], [0.6, 0.7], [0.7, 0.8], [0.8, 0.9], [0.9, 1.0]]
    """
    t = []
    duration = end - start
    step = duration / n
    for i in range(n):
        interval = [start, start + step]
        t.append(interval)
        start = start + step
    return t
